from keckdrpframework.primitives.base_primitive import BasePrimitive
from kcwidrp.primitives.kcwi_file_primitives import kcwi_fits_writer

import numpy as np
import pkg_resources
import os
import pandas as pd


class CorrectDefects(BasePrimitive):
    """
    Remove known bad columns.

    Looks for a defect list file in the data directory of kcwidrp based on the
    CCD ampmode and x and y binning.  Records the defect correction in the
    FITS header with the following keywords:

        * BPFILE: the bad pixel file used to correct defects
        * NBPCLEAN: the number of bad pixels cleaned

    Uses the following configuration parameter:

        * saveintims: if set to ``True`` write out a \*_def.fits file with defects corrected.  Default is ``False``.

    Updates image in returned arguments.

    """

    def __init__(self, action, context):
        BasePrimitive.__init__(self, action, context)
        self.logger = context.pipeline_logger

    def _perform(self):
        self.logger.info("Correcting detector defects")

        # Header keyword to update
        key = 'BPCLEAN'
        keycom = 'cleaned bad pixels?'

        # Create flags for bad columns fixed
        if self.action.args.ccddata.flags is None:
            self.action.args.ccddata.flags = np.zeros(
                self.action.args.ccddata.data.shape, dtype=np.uint8)

        flags = self.action.args.ccddata.flags

        # Nod and Shuffle?
        if self.action.args.nasmask and self.action.args.numopen > 1:
            nastr = "_nas"
        else:
            nastr = ""

        # Does the defect file exist?
        path = "data/defect_%s_%dx%d%s.dat" % (self.action.args.ampmode.strip(),
                                               self.action.args.xbinsize,
                                               self.action.args.ybinsize, nastr)
        package = __name__.split('.')[0]
        full_path = pkg_resources.resource_filename(package, path)
        number_of_bad_pixels = 0   # count of defective pixels cleaned
        if os.path.exists(full_path):
            self.logger.info("Reading defect list in: %s" % full_path)
            defect_table = pd.read_csv(full_path, sep=r'\s+')
            # range of pixels for calculating good value
            pixel_range_for_good_value = 2
            for index, row in defect_table.iterrows():
                # Get coords and adjust for python zero bias
                x0 = row['X0'] - 1
                x1 = row['X1']
                y0 = row['Y0'] - 1
                y1 = row['Y1']
                # Loop over y range
                for by in range(y0, y1):
                    # sample on low side of bad area
                    values = list(self.action.args.ccddata.data[by,
                                  x0-pixel_range_for_good_value:x0])
                    # sample on high side
                    values.extend(self.action.args.ccddata.data[by,
                                  x1:x1+pixel_range_for_good_value])
                    # get replacement value
                    good_values = np.nanmedian(np.asarray(values))
                    # Replace baddies with good_values
                    for bx in range(x0, x1):
                        self.action.args.ccddata.data[by, bx] = good_values
                        flags[by, bx] += 2
                        number_of_bad_pixels += 1
            self.action.args.ccddata.header[key] = (True, keycom)
            self.action.args.ccddata.header['BPFILE'] = (path, 'defect list')
        else:
            self.logger.error("Defect list not found for %s" % full_path)
            self.action.args.ccddata.header[key] = (False, keycom)

        self.logger.info("Cleaned %d bad pixels" % number_of_bad_pixels)
        self.action.args.ccddata.header['NBPCLEAN'] = \
            (number_of_bad_pixels, 'number of bad pixels cleaned')

        log_string = CorrectDefects.__module__
        self.action.args.ccddata.header['HISTORY'] = log_string
        self.logger.info(log_string)

        # add flags array
        # DN 2023-may-28: commenting out because it causes bad things later on
        # self.action.args.ccddata.mask = flags
        self.action.args.ccddata.flags = flags

        if self.config.instrument.saveintims:
            kcwi_fits_writer(self.action.args.ccddata,
                             table=self.action.args.table,
                             output_file=self.action.args.name,
                             output_dir=self.config.instrument.output_directory,
                             suffix="def")

        return self.action.args
    # END: class CorrectDefects()



from keckdrpframework.primitives.base_primitive import BasePrimitive
from kcwidrp.primitives.kcwi_file_primitives import kcwi_fits_writer, \
    strip_fname

import numpy as np

from astropy.nddata import CCDData


class NandshuffSubtractSky(BasePrimitive):
    """
    Locate object and sky panels and perform a nod-and-shuffle sky subtraction.

    With a nod-and-shuffle observation, the resulting image will have a sky and
    an object panel with interleaved exposures of the same total duration.  This
    routine identifies those panels and does a subtraction that should eliminate
    the sky accurately, as long as the interleaving occurs at a rate that tracks
    the changes in sky throughout the exposure.

    Writes out a \*_obj.fits and \*_sky.fits image with the panels aligned, but
    not subtracted.  These are subsequently processed through the pipeline to
    allow the confirmation that resulting features are not the result of sky
    subtraction.

    Also writes out the sky-subtracted image in a \*_intk.fits image and adds
    an entry in the proc table.

    """

    def __init__(self, action, context):
        BasePrimitive.__init__(self, action, context)
        self.logger = context.pipeline_logger

    def _pre_condition(self):
        """
        Checks if it is a nod-and-shuffle observation
        """
        self.logger.info("Checking precondition for NandshuffSubtractSky")
        if self.action.args.nasmask and self.action.args.numopen > 1:
            self.logger.info("Preconditions for NandshuffSubtractSky met.")
            return True
        else:
            self.logger.warning("Precondition not met: "
                                "not a nod-and-shuffle observation.")
            return False

    def _perform(self):
        self.logger.info("Subtracting nod-and shuffle sky background")

        # Header keyword to update
        key = 'NASSUB'
        keycom = 'Nod-and-shuffle subtraction done?'
        target_type = 'SKY'

        # get header values
        ofn = self.action.args.name
        shrows = self.action.args.ccddata.header['SHUFROWS']
        nshfup = self.action.args.ccddata.header['NSHFUP']
        nshfdn = self.action.args.ccddata.header['NSHFDN']
        camera = self.action.args.ccddata.header['CAMERA'].upper()
        ybin = int(self.action.args.ccddata.header['BINNING'].split(',')[1])

        # units
        u_out = self.action.args.ccddata.unit

        # BLUE camera handling
        if 'BLUE' in camera:
            # BLUE nominal conditions (sky on bottom, object in middle)
            skyrow0 = 0
            skyrow1 = shrows - 1
            objrow0 = shrows
            objrow1 = shrows + shrows - 1
            # record subtraction rows
            subrow0 = objrow0
            subrow1 = objrow1

            # aborted script with inverted panels (sky in middle, object above)
            if nshfdn != nshfup + 1:
                self.logger.warning("Inverted N&S obs detected!")
                skyrow0 = shrows
                skyrow1 = shrows + shrows - 1
                objrow0 = skyrow1 + 1
                objrow1 = objrow0 + shrows - 1
                # record subtraction rows
                subrow0 = skyrow0
                subrow1 = skyrow1

        # RED camera handling
        elif 'RED' in camera:
            # RED uses unbinned pixels for shrows
            if ybin == 2:
                shrows = int(shrows / 2) + 1
            # RED nominal conditions (object on bottom, sky in middle)
            objrow0 = 0
            objrow1 = shrows - 1
            skyrow0 = objrow1 + 1
            skyrow1 = skyrow0 + shrows - 1
            # record subtraction rows
            subrow0 = skyrow0
            subrow1 = skyrow1

            # aborted script with inverted panels (object in middle sky above)
            if nshfdn != nshfup:
                self.logger.warning("Inverted N&S obs detected!")
                objrow0 = shrows
                objrow1 = objrow0 + shrows - 1
                skyrow0 = objrow1 + 1
                skyrow1 = skyrow0 + shrows - 1
                # record subtraction rows
                subrow0 = objrow0
                subrow1 = objrow1

        else:
            self.logger.error("CAMERA cannot be determined for N&S Sub")
            return self.action.args

        # check limits
        self.logger.info(
            "Nod-and-shuffle rows (sky0, 1, obj0,1): %d, %d, %d, %d" %
            (skyrow0, skyrow1, objrow0, objrow1))
        if (skyrow1-skyrow0) != (objrow1-objrow0):
            self.logger.error("Nod-and-shuffle panel mis-match error")
            return self.action.args
        else:
            self.logger.info("Nod-and-shuffle panels match")

        # record in header (and convert to 1-bias, for IRAF)
        self.action.args.ccddata.header['OBJROW0'] = (objrow0+1, "Object row 0")
        self.action.args.ccddata.header['OBJROW1'] = (objrow1+1, "Object row 1")
        self.action.args.ccddata.header['SKYROW0'] = (skyrow0+1, "Sky row 0")
        self.action.args.ccddata.header['SKYROW1'] = (skyrow1+1, "Sky row 1")
        self.action.args.ccddata.header['SUBROW0'] = (subrow0 + 1,
                                                      "Subtraction row 1")
        self.action.args.ccddata.header['SUBROW1'] = (subrow1 + 1,
                                                      "Subtraction row 1")

        # create intermediate images and headers
        sky = self.action.args.ccddata.data.copy()
        obj = self.action.args.ccddata.data.copy()
        std = self.action.args.ccddata.uncertainty.array.copy()
        msk = self.action.args.ccddata.mask.copy()
        flg = self.action.args.ccddata.flags.copy()
        skyhdr = self.action.args.ccddata.header.copy()
        objhdr = self.action.args.ccddata.header.copy()

        # BLUE camera handling
        if 'BLUE' in camera:
            # nominal condition
            if skyrow0 < 10:
                self.logger.info("standard nod-and-shuffle configuration")
                skystd = self.action.args.ccddata.uncertainty.array.copy()
                skymsk = self.action.args.ccddata.mask.copy()
                skyflg = self.action.args.ccddata.flags.copy()
                # move sky to object position
                sky[objrow0:objrow1+1, :] = obj[skyrow0:skyrow1+1, :]
                skystd[objrow0:objrow1+1, :] = std[skyrow0:skyrow1+1, :]
                skymsk[objrow0:objrow1+1, :] = msk[skyrow0:skyrow1+1, :]
                skyflg[objrow0:objrow1+1, :] = flg[skyrow0:skyrow1+1, :]
                # do subtraction
                self.action.args.ccddata.data -= sky
                self.action.args.ccddata.uncertainty.array = np.sqrt(
                    std ** 2 + skystd ** 2)
                self.action.args.ccddata.mask += skymsk
                self.action.args.ccddata.flags |= skyflg
                # clean images
                self.action.args.ccddata.data[skyrow0:skyrow1+1, :] = 0.
                self.action.args.ccddata.data[objrow1+1:-1, :] = 0.
                self.action.args.ccddata.uncertainty.array[
                    skyrow0:skyrow1+1, :] = 0.
                self.action.args.ccddata.uncertainty.array[objrow1+1:-1, :] = 0.
                self.action.args.ccddata.mask[skyrow0:skyrow1+1, :] = 1
                self.action.args.ccddata.mask[objrow1+1:-1, :] = 1
                self.action.args.ccddata.flags[skyrow0:skyrow1+1, :] = 64
                self.action.args.ccddata.flags[objrow1+1:-1, :] = 64
                sky[skyrow0:skyrow1+1, :] = 0.
                sky[objrow1+1:-1, :] = 0.
                obj[skyrow0:skyrow1+1, :] = 0.
                obj[objrow1+1:-1, :] = 0.
            else:
                self.logger.warning(
                    "non-standard nod-and-shuffle configuration")
                skyscl = -1.
                while skyscl < 0.:
                    if self.config.instrument.plot_level >= 2:
                        q = input("Enter scale factor for sky to match obj "
                                  "(float): ")
                        try:
                            skyscl = float(q)
                        except ValueError:
                            self.logger.warning(
                                "Invalid input: %s, try again" % q)
                            skyscl = -1.0
                    else:
                        skyscl = 1.0
                self.logger.info("Sky scaling used = %.2f" % skyscl)
                objstd = self.action.args.ccddata.uncertainty.array.copy()
                objmsk = self.action.args.ccddata.mask.copy()
                objflg = self.action.args.ccddata.flags.copy()
                # move object to sky position
                obj[skyrow0:skyrow1+1, :] = obj[objrow0:objrow1+1, :]
                objstd[skyrow0:skyrow1+1, :] = std[objrow0:objrow1+1, :]
                objmsk[skyrow0:skyrow1+1, :] = msk[objrow0:objrow1+1, :]
                objflg[skyrow0:skyrow1+1, :] = flg[objrow0:objrow1+1, :]
                # do subtraction
                sky *= skyscl
                self.action.args.ccddata.data = obj - sky
                self.action.args.ccddata.uncertainty.array = np.sqrt(
                    std ** 2 + objstd ** 2)
                self.action.args.ccddata.mask += objmsk
                self.action.args.ccddata.flags |= objflg
                # clean images
                self.action.args.ccddata.data[objrow0:-1, :] = 0.
                self.action.args.ccddata.data[0:skyrow0+1, :] = 0.
                self.action.args.ccddata.uncertainty.array[objrow0:-1, :] = 0.
                self.action.args.ccddata.uncertainty.array[0:skyrow0+1, :] = 0.
                self.action.args.ccddata.mask[objrow0:-1, :] = 1
                self.action.args.ccddata.mask[0:skyrow0+1, :] = 1
                self.action.args.ccddata.flags[objrow0:-1, :] = 64
                self.action.args.ccddata.flags[0:skyrow0+1, :] = 64
                sky[objrow0:-1, :] = 0.
                sky[0:skyrow0+1, :] = 0.
                obj[objrow0:-1, :] = 0.
                obj[0:skyrow0+1, :] = 0.
                cmnt = 'Aborted nod-and-shuffle observations'
                objhdr['COMMENT'] = cmnt
                skyhdr['COMMENT'] = cmnt
                self.action.args.ccddata.header['COMMENT'] = cmnt
                skyhdr['NASSCL'] = (skyscl, 'Scale factor applied to sky panel')
                self.action.args.ccddata.header['NASSCL'] = (
                    skyscl, 'Scale factor applied to sky panel')
        elif 'RED' in camera:
            # nominal condition
            if objrow0 < 10:
                self.logger.info("standard nod-and-shuffle configuration")
                objstd = self.action.args.ccddata.uncertainty.array.copy()
                objmsk = self.action.args.ccddata.mask.copy()
                objflg = self.action.args.ccddata.flags.copy()
                # move object to sky position
                obj[skyrow0:skyrow1+1, :] = obj[objrow0:objrow1+1, :]
                objstd[objrow0:objrow1+1, :] = std[skyrow0:skyrow1+1, :]
                objmsk[objrow0:objrow1+1, :] = msk[skyrow0:skyrow1+1, :]
                objflg[objrow0:objrow1+1, :] = flg[skyrow0:skyrow1+1, :]
                # do subtraction
                self.action.args.ccddata.data = obj - sky
                self.action.args.ccddata.uncertainty.array = np.sqrt(
                    std ** 2 + objstd ** 2)
                self.action.args.ccddata.mask += objmsk
                self.action.args.ccddata.flags |= objflg
                # clean images
                self.action.args.ccddata.data[objrow0:objrow1+1, :] = 0.
                self.action.args.ccddata.data[skyrow1+1:-1, :] = 0.
                self.action.args.ccddata.uncertainty.array[
                    objrow0:objrow1+1, :] = 0.
                self.action.args.ccddata.uncertainty.array[skyrow1+1:-1, :] = 0.
                self.action.args.ccddata.mask[objrow0:objrow1+1, :] = 1
                self.action.args.ccddata.mask[skyrow1+1:-1, :] = 1
                self.action.args.ccddata.flags[objrow0:objrow1+1, :] = 64
                self.action.args.ccddata.flags[skyrow1+1:-1, :] = 64
                obj[objrow0:objrow1+1, :] = 0.
                obj[skyrow1+1:-1, :] = 0.
                sky[objrow0:objrow1+1, :] = 0.
                sky[skyrow1+1:-1, :] = 0.
            else:
                self.logger.warning(
                    "non-standard nod-and-shuffle configuration")
                skyscl = -1.
                while skyscl < 0.:
                    if self.config.instrument.plot_level >= 2:
                        q = input("Enter scale factor for sky to match obj "
                                  "(float): ")
                        try:
                            skyscl = float(q)
                        except ValueError:
                            self.logger.warning(
                                "Invalid input: %s, try again" % q)
                            skyscl = -1.0
                    else:
                        skyscl = 1.0
                self.logger.info("Sky scaling used = %.2f" % skyscl)
                skystd = self.action.args.ccddata.uncertainty.array.copy()
                skymsk = self.action.args.ccddata.mask.copy()
                skyflg = self.action.args.ccddata.flags.copy()
                # move sky to object position
                sky[objrow0:objrow1+1, :] = obj[skyrow0:skyrow1+1, :]
                skystd[objrow0:objrow1+1, :] = std[skyrow0:skyrow1+1, :]
                skymsk[objrow0:objrow1+1, :] = msk[skyrow0:skyrow1+1, :]
                skyflg[objrow0:objrow1+1, :] = flg[skyrow0:skyrow1+1, :]
                # do subtraction
                self.action.args.ccddata.data -= sky
                self.action.args.ccddata.uncertainty.array = np.sqrt(
                    std ** 2 + skystd ** 2)
                self.action.args.ccddata.mask += skymsk
                self.action.args.ccddata.flags |= skyflg
                # clean images
                self.action.args.ccddata.data[0:objrow0, :] = 0.
                self.action.args.ccddata.data[objrow1+1:-1, :] = 0.
                self.action.args.ccddata.uncertainty.array[0:objrow0, :] = 0.
                self.action.args.ccddata.uncertainty.array[objrow1+1:-1:] = 0.
                self.action.args.ccddata.mask[0:objrow0, :] = 1
                self.action.args.ccddata.mask[objrow1+1:-1, :] = 1
                self.action.args.ccddata.flags[0:objrow0, :] = 64
                self.action.args.ccddata.flags[objrow1+1:-1, :] = 64
                sky[0, objrow0, :] = 0.
                sky[objrow1+1:-1, :] = 0.
                obj[0, objrow0, :] = 0.
                obj[objrow1+1:-1, :] = 0.
                cmnt = 'Aborted nod-and-shuffle observations'
                objhdr['COMMENT'] = cmnt
                skyhdr['COMMENT'] = cmnt
                self.action.args.ccddata.header['COMMENT'] = cmnt
                skyhdr['NASSCL'] = (skyscl, 'Scale factor applied to sky panel')
                self.action.args.ccddata.header['NASSCL'] = (
                    skyscl, 'Scale factor applied to sky panel')
        else:
            self.logger.error("CAMERA cannot be determined for N&S Sub")
            return self.action.arg

        # log
        self.logger.info("Nod-and-shuffle subtracted")
        # update headers
        log_string = NandshuffSubtractSky.__module__

        objhdr[key] = (False, keycom)
        objhdr['HISTORY'] = log_string

        skyhdr[key] = (False, keycom)
        skyhdr['SKYOBS'] = (True, 'Sky observation?')
        skyhdr['HISTORY'] = log_string

        self.action.args.ccddata.header[key] = (True, keycom)
        self.action.args.ccddata.header['HISTORY'] = log_string

        # write out sky image
        msname = strip_fname(ofn) + '_' + target_type.lower() + '.fits'
        out_sky = CCDData(sky, meta=skyhdr, unit=u_out)
        kcwi_fits_writer(out_sky, output_file=msname,
                         output_dir=self.config.instrument.output_directory)
        # write out object image
        obname = strip_fname(ofn) + '_obj.fits'
        out_obj = CCDData(obj, meta=objhdr, unit=u_out)
        kcwi_fits_writer(out_obj, output_file=obname,
                         output_dir=self.config.instrument.output_directory)

        # update header keywords
        self.action.args.ccddata.header['SKYMAST'] = (msname,
                                                      "Master sky filename")

        # write out int image
        kcwi_fits_writer(self.action.args.ccddata,
                         table=self.action.args.table,
                         output_file=self.action.args.name,
                         output_dir=self.config.instrument.output_directory,
                         suffix="intk")
        self.context.proctab.update_proctab(frame=self.action.args.ccddata,
                                            suffix="intk",
                                            filename=self.action.args.name)
        self.context.proctab.write_proctab(tfil=self.config.instrument.procfile)

        self.logger.info(log_string)

        return self.action.args
    # END: class NandshuffSubtractSky()



from keckdrpframework.primitives.base_primitive import BasePrimitive
from kcwidrp.core.bokeh_plotting import bokeh_plot
from kcwidrp.core.kcwi_plotting import save_plot
from kcwidrp.primitives.kcwi_file_primitives import plotlabel

from bokeh.plotting import figure
from bokeh.models import Range1d
import pkg_resources
import os
from astropy.io import fits as pf
import numpy as np
from scipy.ndimage import gaussian_filter1d
from scipy.interpolate import interpolate
from scipy import signal
import time


class ReadAtlas(BasePrimitive):
    """
    Read in atlas spectrum and derive alignment offset.

    Reads in spectral atlas spectrum of observed arc lamp.  Uses instrument
    configuration to determine convoution kernel that will match the observed
    spectrum resolution and applies it.  Then picks a central part of the
    observed spectrum and tapers the ends and does a cross-correlation to
    determine the pixel offset that will align the observed and atlas spectra.

    Uses the following configuration parameters:

        * MIDFRAC: what central fraction to use for finding the atlas offset - defaults to 0.3.
        * TAPERFRAC: how much of the ends to taper off (in case of bright lines at the ends) - defaults to 0.2.
        * ATOFF: force a specific pixel offset for the atlas - defaults to 0.

    Places convolved atlas spectrum with wavelength scale in the return
    argument so subsequent processing can use it to fit the observed wavelength
    scale accurately.

    """

    def __init__(self, action, context):
        BasePrimitive.__init__(self, action, context)
        self.logger = context.pipeline_logger

    def _pre_condition(self):
        self.logger.info("Checking for master arc")
        if 'MARC' in self.action.args.ccddata.header['IMTYPE']:
            return True
        else:
            return False

    def _perform(self):
        # What lamp are we using?
        lamp = self.action.args.illum
        # Does the atlas file exist?
        path = "data/%s.fits" % lamp.lower()  # always use slash
        pkg = __name__.split('.')[0]
        atpath = pkg_resources.resource_filename(pkg, path)
        if os.path.exists(atpath):
            self.logger.info("Reading atlas spectrum in: %s" % atpath)
        else:
            self.logger.error("Atlas spectrum not found for %s" % atpath)
        # Read the atlas
        ff = pf.open(atpath)
        reflux = ff[0].data
        refdisp = ff[0].header['CDELT1']
        refwav = np.arange(0, len(reflux)) * refdisp + ff[0].header['CRVAL1']
        ff.close()
        # Convolve with appropriate Gaussian
        self.logger.info("Convolving Atlas with Gaussian having sigma of %.2f "
                         "px" % self.action.args.atsig)
        reflux = gaussian_filter1d(reflux, self.action.args.atsig)
        # Observed arc spectrum
        obsarc = self.context.arcs[self.config.instrument.REFBAR]
        # Preliminary wavelength solution
        xvals = np.arange(0, len(obsarc)) - int(len(obsarc)/2)
        obswav = xvals * self.context.prelim_disp + self.action.args.cwave
        # Define Central Region
        mf = self.config.instrument.MIDFRAC
        # Set on command line
        if 0 < mf <= 1.0:
            self.logger.info("Using command line value for middle fraction")
        # Default values based on grating
        else:
            self.logger.info("Using default calculated middle fraction")
            # Get central third
            mf = 1./3.
            # Unless we are low disp., then get central 3 5ths
            if 'BL' in self.action.args.grating or \
                    'RL' in self.action.args.grating:
                mf = 3./5.
            # Unless we are red high disp., then get central 8 10ths
            if 'RH' in self.action.args.grating:
                mf = 8./10.
        self.logger.info("Central %.3f used for initial fit" % mf)
        minow = int(len(obsarc) * (0.5 - mf / 2.))
        maxow = int(len(obsarc) * (0.5 + mf / 2.))
        if self.context.prelim_disp > 0:
            minwav = obswav[minow]
            maxwav = obswav[maxow]
        else:
            minwav = obswav[maxow]
            maxwav = obswav[minow]
        # Check for dichroic
        if self.action.args.dich:
            obs_extent = maxow - minow
            if self.action.args.camera == 0:  # Blue
                # test if correction needed
                if maxwav > 5600:
                    maxwav = [w for w in obswav if w < 5620.][-1]
                    maxow = [i for i, w in enumerate(obswav) if w < 5620.][-1]
                    minow = maxow - obs_extent
                    if minow < 0:
                        minow = 0
                    self.logger.info("Dichroic adjusted - x0, x1, w0, w1: "
                                     "%d, %d, %.2f, %.2f" % (minow, maxow,
                                                             minwav, maxwav))
            elif self.action.args.camera == 1:  # Red
                # test if correction needed
                if minwav < 5600:
                    minwav = [w for w in obswav if w > 5580.][0]
                    minow = [i for i, w in enumerate(obswav) if w > 5580.][0]
                    maxow = minow + obs_extent
                    if maxow > (len(obsarc) - 1):
                        maxow = len(obsarc) - 1
                    self.logger.info("Dichroic adjusted - x0, x1, w0, w1: "
                                     "%d, %d, %.2f, %.2f" % (minow, maxow,
                                                             minwav, maxwav))
            else:
                self.logger.warning("Camera undefined!!")
        # Get corresponding ref range
        minrw = [i for i, v in enumerate(refwav) if v >= minwav][0]
        maxrw = [i for i, v in enumerate(refwav) if v <= maxwav][-1]
        # Subsample for cross-correlation
        cc_obsarc = obsarc[minow:maxow]
        cc_obswav = obswav[minow:maxow]
        cc_reflux = reflux[minrw:maxrw]
        cc_refwav = refwav[minrw:maxrw]
        # Resample onto reference wavelength scale
        obsint = interpolate.interp1d(cc_obswav, cc_obsarc, kind='cubic',
                                      bounds_error=False,
                                      fill_value='extrapolate')
        cc_obsarc = obsint(cc_refwav)
        # Apply cosin bell taper to both
        cc_obsarc *= signal.windows.tukey(
            len(cc_obsarc), alpha=self.config.instrument.TAPERFRAC)
        cc_reflux *= signal.windows.tukey(
            len(cc_reflux), alpha=self.config.instrument.TAPERFRAC)
        nsamp = len(cc_refwav)
        offar = np.arange(1 - nsamp, nsamp)
        # Cross-correlate
        xcorr = np.correlate(cc_obsarc, cc_reflux, mode='full')
        # Get central region
        x0c = int(len(xcorr)/3)
        x1c = int(2*(len(xcorr)/3))
        xcorr_central = xcorr[x0c:x1c]
        offar_central = offar[x0c:x1c]
        # Calculate offset
        calc_offset_pix = offar_central[xcorr_central.argmax()]
        calc_offset_wav = calc_offset_pix * refdisp
        self.logger.info("Calculated arc-atlas offset (px, Ang): %d, %.1f" %
                         (calc_offset_pix, calc_offset_wav))
        atoff = self.config.instrument.ATOFF
        req_offset_pix = 0
        if atoff != 0:
            req_offset_pix = atoff
            req_offset_wav = atoff * refdisp
            self.logger.info("Command line requested offset (px, Ang):"
                             " %d, %.1f" % (req_offset_pix, req_offset_wav))
            offset_pix = req_offset_pix
            offset_wav = req_offset_wav
        else:
            offset_pix = calc_offset_pix
            offset_wav = calc_offset_wav
        plab = plotlabel(self.action.args)
        if self.config.instrument.plot_level >= 1:
            # Plot
            p = figure(title=plab + "ATLAS OFFSET = %d px" % offset_pix,
                       x_axis_label="Offset(px)", y_axis_label="X-corr",
                       plot_width=self.config.instrument.plot_width,
                       plot_height=self.config.instrument.plot_height)

            p.line(offar_central, xcorr_central, legend_label='Data')
            ylim_min = min(xcorr_central)
            ylim_max = max(xcorr_central)
            p.line([calc_offset_pix, calc_offset_pix], [ylim_min, ylim_max],
                   color='red', legend_label='Peak')
            if atoff != 0:
                p.line([req_offset_pix, req_offset_pix], [ylim_min, ylim_max],
                       color='green', legend_label='Reqested')
            bokeh_plot(p, self.context.bokeh_session)
            if self.config.instrument.plot_level >= 2:
                input("Next? <cr>: ")
            else:
                time.sleep(self.config.instrument.plot_pause)
            # Get central wavelength
            cwave = self.action.args.cwave
            # Set up offset tweaking
            q = 'test'
            while q:
                # Plot the two spectra
                p = figure(title=plab + "ATLAS OFFSET = %.1f Ang (%d px)" %
                           (offset_wav, offset_pix),
                           x_axis_label="Wave(A)", y_axis_label="Rel. Flux",
                           plot_width=self.config.instrument.plot_width,
                           plot_height=self.config.instrument.plot_height)
                p.line(obswav[minow:maxow] - offset_wav,
                       obsarc[minow:maxow]/np.nanmax(obsarc[minow:maxow]),
                       legend_label="ref bar (%d)" %
                       self.config.instrument.REFBAR)
                p.line(refwav[minrw:maxrw],
                       reflux[minrw:maxrw]/np.nanmax(reflux[minrw:maxrw]),
                       color="red", legend_label="Atlas")
                p.x_range = Range1d(minwav, maxwav)
                ylim_min = min(
                    obsarc[minow:maxow]/np.nanmax(obsarc[minow:maxow]))
                ylim_max = max(
                    obsarc[minow:maxow]/np.nanmax(obsarc[minow:maxow]))
                p.line([cwave, cwave], [ylim_min, ylim_max], color="magenta",
                       legend_label="CWAVE", line_dash="dashdot")
                bokeh_plot(p, self.context.bokeh_session)

                if self.config.instrument.plot_level >= 2:
                    q = input("Enter: <cr> - next, new offset (int px): ")
                    if q:
                        try:
                            offset_pix = int(q)
                            offset_wav = offset_pix * refdisp
                        except ValueError:
                            print("Try again: integer pixel values accepted")
                            q = 'test'
                else:
                    time.sleep(self.config.instrument.plot_pause)
                    q = None
            # save final plot of arc/atlas offset
            pfname = "arc_%05d_%s_%s_%s" % (
                self.action.args.ccddata.header['FRAMENO'],
                self.action.args.illum, self.action.args.grating,
                self.action.args.ifuname)
            save_plot(p, filename=pfname + '_atoff.png')
            self.logger.info("Final   arc-atlas offset (px, Ang): %d, %.1f" %
                             (offset_pix, offset_wav))
        # Store atlas spectrum
        self.action.args.reflux = reflux
        self.action.args.refwave = refwav
        # Store offsets
        self.action.args.offset_pix = offset_pix
        self.action.args.offset_wave = offset_wav
        # Store reference dispersion
        self.action.args.refdisp = refdisp
        # Store central limits
        self.action.args.minrow = minow
        self.action.args.maxrow = maxow
        # Store x values
        self.action.args.xvals = xvals
        self.action.args.x0 = int(len(obsarc)/2)

        log_string = ReadAtlas.__module__
        self.action.args.ccddata.header['HISTORY'] = log_string
        self.logger.info(log_string)

        return self.action.args
    # END: class ReadAtlas()



import astropy.utils.exceptions as aue


class PydlutilsException(Exception):
    """Base class for exceptions raised in PyDL functions.
    """
    pass


class PydlutilsUserWarning(aue.AstropyUserWarning):
    """Class for warnings issued by :mod:`pydl.pydlutils`.
    """
    pass


__all__ = ['PydlutilsException', 'PydlutilsUserWarning']


#!/usr/bin/env python
import sys
if sys.version_info[0] > 2:
    from astropy.io import fits as pf
else:
    import pyfits as pf


def wb_main():
    """
    Generate summary log and group list files for BLUE channel images.

    Call get_log_string to create summary entry for each image and group them
    according to processing group.  Write out unique processing group lists
    in \*.txt files.  These files can be input to the pipeline with the -l
    command line parameter to allow processing of groups one at a time.  For
    example, 2x2 Blue biases taken with the TUP amp configuration in slow
    readout with gainmul 10 will end up in the file bias2x2TUP010_0.txt.  A
    master bias can be created by issuing the following command:

        >>> reduce_kcwi -b -l bias2x2TUP010_0.txt

    These group files are generated for biases, darks, continuum bars, arcs,
    flats, and all objects.  The filenames are all appended with the last four
    characters of the STATEID header keyword, so identical configurations from
    different states can be distinguished.

    Always good to type out the list file before processing it.

    Examples:

        >>> wb kb*.fits > whatb.list

        This will generate a summary log file along with associated group list
        files that can be used as inputs to the `reduce_kcwi` command with the
        `-l` parameter.  An example of the resulting \*.txt files is below::

            SN2023ixf2x2MedKBlueBL4500_75fe.txt      bias2x2TUP010_0.txt
            allb.txt                                 cbars2x2MedKBlueBL_4500_0.7_75fe.txt
            arcs2x2MedKBlueBLFeAr4500_10.0_75fe.txt  cflat2x2MedKBlueBL_4500_0.7_75fe.txt
            arcs2x2MedKBlueBLThAr4500_20.0_75fe.txt  dflat2x2MedKBlueBL_4500_14.0_75fe.txt
            bd26d26062x2MedKBlueBL4500_75fe.txt

        One can proceed through processing steps like this:

        >>> reduce_kcwi -b -l bias2x2TUP010_0.txt
        >>> reduce_kcwi -b -l cbars2x2MedKBlueBL_4500_0.7_75fe.txt
        >>> reduce_kcwi -b -l arcs2x2MedKBlueBLThAr4500_20.0_75fe.txt
        >>> reduce_kcwi -b -l cflat2x2MedKBlueBL_4500_0.7_75fe.txt
        >>> reduce_kcwi -b -l bd26d26062x2MedKBlueBL4500_75fe.txt
        >>> reduce_kcwi -b -l SN2023ixf2x2MedKBlueBL4500_75fe.txt

    """
    import sys

    if len(sys.argv) < 2:
        print("Usage - wb <fspec>")
    else:
        configs = []
        fnames = {"allb": []}
        for ifl in sys.argv[1:]:
            logstr, cfgstr, lsfn = get_log_string(ifl, batch=True)
            print(logstr)
            fnames['allb'].append(ifl)
            if lsfn:
                if lsfn in fnames:
                    fnames[lsfn].append(ifl)
                else:
                    fnames[lsfn] = [ifl]
            if cfgstr:
                configs.append(cfgstr)
        # Unique configs
        uconfigs = sorted(set(configs))
        print("Number of unique configurations = %d" % len(uconfigs))
        for c in uconfigs:
            print(c)

        for cal in fnames:
            with open(cal+".txt", 'w') as cal_list:
                for f in fnames[cal]:
                    cal_list.write(f + "\n")


def get_cal_list_file(hdr):
    """
    Return list file name given configuration in header.

    Generates the group file name apropriate for the input image header.

    Args:
        hdr (FITS header): the current image header

    :returns:
        (str): group filename

    :meta private:
    """

    fpre = {"ARCLAMP": "arcs", "CONTBARS": "cbars", "FLATLAMP": "cflat",
            "DOMEFLAT": "dflat", "TWIFLAT": "tflat"}

    imtype = hdr['IMTYPE']
    lfname = None
    if imtype in fpre:
        lfname = fpre[imtype]
        lfname += hdr['BINNING'].replace(',', 'x')
        lfname += hdr['IFUNAM'][:3]
        lfname += hdr['BFILTNAM']
        lfname += hdr['BGRATNAM']
        if 'arcs' in lfname:
            if 'FeAr' in hdr['ILLUME']:
                lfname += 'FeAr'
            else:
                lfname += 'ThAr'
        else:
            lfname += '_'
        lfname += "%.0f" % hdr['BCWAVE']
        lfname += "_%.1f" % hdr['EXPTIME']
        lfname += "_" + hdr['CONFIGID']
    elif 'BIAS' in imtype:
        lfname = 'bias'
        lfname += hdr['BINNING'].replace(',', 'x')
        lfname += hdr['AMPMODE']
        lfname += str(hdr['CCDMODE'])
        lfname += str(hdr['GAINMUL'])
        lfname += "_" + hdr['CONFIGID']
    elif 'DARK' in imtype:
        lfname = 'dark'
        lfname += hdr['BINNING'].replace(',', 'x')
        lfname += hdr['AMPMODE']
        lfname += str(hdr['CCDMODE'])
        lfname += str(hdr['GAINMUL'])
        lfname += "_%.1f" % hdr['EXPTIME']
        lfname += "_" + hdr['CONFIGID']
    elif 'OBJECT' in imtype:
        lfname = hdr['TARGNAME']
        lfname += hdr['BINNING'].replace(',', 'x')
        lfname += hdr['IFUNAM'][:3]
        lfname += hdr['BFILTNAM']
        lfname += hdr['BGRATNAM']
        lfname += "%.0f" % hdr['BCWAVE']
        lfname += "_" + hdr['CONFIGID']
    return lfname


def get_log_string(ifile, batch=False):
    """
    Generate log entry from BLUE FITS header keywords.

    Attempt to encapsulate the instrument configuration for each image by
    summarizing FITS header keyword values tersely.

    Args:
        ifile (str): filename of FITS image to summarize
        batch (bool): set to ``True`` for an abreviated record.  Defaults to ``False``.

    :returns:
        (str): Configuration summary string for the input FITS image file.

    :meta private:
    """

    try:
        ff = pf.open(ifile)
    except IOError:
        print("***ERROR*** empty or corrupt fits file: %s" % ifile)
        return None, None

    header = ff[0].header
    header['FNAME'] = ifile
    if 'CAMERA' in header:
        if 'BLUE' in header['CAMERA'].upper():
            is_bias = False
            if 'OFNAME' not in header:
                header['OFNAME'] = ifile
            if 'STATEID' not in header:
                header['CONFIGID'] = '----'
            else:
                header['CONFIGID'] = header['STATEID'][-4:]
            if 'AMPMODE' not in header:
                header['AMPMODE'] = '-'
            if 'BINNING' not in header:
                header['BINNING'] = '-'
            if 'CCDMODE' not in header:
                header['CCDMODE'] = -1
            if 'GAINMUL' not in header:
                header['GAINMUL'] = -1
            if 'NUMOPEN' not in header:
                header['NUMOPEN'] = -1
            if 'XPOSURE' not in header:
                header['XPOSURE'] = -1.
            if 'AIRMASS' not in header:
                header['AIRMASS'] = -1.
            if 'TELAPSE' not in header:
                header['TELAPSE'] = -1.
            if 'BFILTNAM' not in header:
                header['BFILTNAM'] = '-'
            if 'BGRATNAM' not in header:
                header['BGRATNAM'] = '-'
            if 'BGROTNAM' not in header:
                header['BGROTNAM'] = '-'
            if 'BCWAVE' not in header:
                header['BCWAVE'] = -1.
            if 'CALMNAM' not in header:
                header['CALMNAM'] = '-'
            if 'CALPNAM' not in header:
                header['CALPNAM'] = '-'
            if 'CALLANG' not in header:
                header['CALLANG'] = -1.
            if 'BARTANG' not in header:
                header['BARTANG'] = -1.
            if 'BNASNAM' not in header:
                header['BNASNAM'] = '-'
            if 'BFOCMM' not in header:
                header['BFOCMM'] = -1.
            if 'CALTYPE' not in header:
                header['CALTYPE'] = '-'
            if 'IFUNAM' not in header:
                header['IFUNAM'] = '-'
            else:
                header['IFUNAM'] = header['IFUNAM'][:3]
            if 'OBJECT' not in header:
                header['OBJECT'] = '-'
            if 'TARGNAME' not in header:
                header['TARGNAME'] = '-'
            if 'CALXNAM' not in header:
                header['CALXNAM'] = '-'
            if 'object' not in header['CALTYPE']:
                header['OBJECT'] = header['CALXNAM']
            if header['TELAPSE'] > header['XPOSURE']:
                header['EXPTIME'] = header['TELAPSE']
            else:
                header['EXPTIME'] = header['XPOSURE']
            if header['EXPTIME'] <= 0.:
                is_bias = True
            header['ILLUME'] = '-'
            try:
                if header['LMP0STAT'] == 1:
                    if header['LMP0SHST'] == 1:
                        header['ILLUME'] = header['LMP0NAM']
                if header['LMP1STAT'] == 1:
                    if header['LMP1SHST'] == 1:
                        header['ILLUME'] = header['LMP1NAM']
                if header['LMP2STAT'] == 1:
                    if header['LMP2SHST'] == 1:
                        header['ILLUME'] = header['LMP2NAM']
                if header['LMP3STAT'] == 1:
                    header['ILLUME'] = header['LMP3NAM'][0:6]
                if 'object' in header['CALTYPE']:
                    if 'on' in header['FLSPECTR'] or 'on' in header['FLIMAGIN']:
                        header['ILLUME'] = 'DOME'
                        if not batch:
                            header['OBJECT'] = 'DOME'
                if 'BIAS' in header['IMTYPE']:
                    if not is_bias:
                        header['IMTYPE'] = 'DARK'
                if 'DARK' in header['IMTYPE']:
                    if is_bias:
                        header['IMTYPE'] = 'BIAS'
                if not batch:
                    if 'object' not in header['CALTYPE']:
                        header['OBJECT'] = header['OBJECT'] + header['ILLUME']
            except KeyError:
                pass
            try:
                lstring = "%(OFNAME)19s %(CONFIGID)4s (%(AMPMODE)3s/%(BINNING)3s/%(CCDMODE)1d/" \
                          "%(GAINMUL)2d/%(NUMOPEN)2d/%(EXPTIME)6.1f s), (%(IFUNAM)3s/" \
                          "%(BFILTNAM)5s/%(BGRATNAM)4s/%(BGROTNAM)9s dg/" \
                          "%(BCWAVE)6.1f/%(CALMNAM)s/%(CALPNAM)5s/%(CALLANG)5.1f dg), " \
                          "(%(BARTANG)5.1f/%(BNASNAM)4s/%(BFOCMM)6.3f) %(AIRMASS)5.3f: %(IMTYPE)7s/" \
                          "%(ILLUME)6s/%(TARGNAME)s:%(OBJECT)s" % header
            except:
                lstring = "%19s : ?" % ifile

            if header['EXPTIME'] <= 0.0:
                cstr = "%(BINNING)3s:%(AMPMODE)3s:%(CCDMODE)1d:%(GAINMUL)2d:BIAS" % \
                       header
            else:
                if batch:
                    cstr = "%(BINNING)3s:%(BFILTNAM)s:%(BGRATNAM)s:%(IFUNAM)s:%(BCWAVE).1f" \
                           % header
                else:
                    cstr = "%(BINNING)3s:%(BFILTNAM)s:%(BGRATNAM)s:%(IFUNAM)s:%(BCWAVE).1f:" \
                           "%(EXPTIME)6.1f:%(OBJECT)s" % header
            lfn = get_cal_list_file(header)
        else:
            lstring = "%19s : NOT a BLUE image!" % ifile
            cstr = None
            lfn = None

    else:
        if not batch:
            print("ERROR - Camera can not be determined.")
        lstring = "%19s : No CAMERA keyword!" % ifile
        cstr = None
        lfn = None

    return lstring, cstr, lfn


if __name__ == '__main__':
    wb_main()



from keckdrpframework.primitives.base_primitive import BasePrimitive
from kcwidrp.primitives.kcwi_file_primitives import parse_imsec, \
    kcwi_fits_writer


class CorrectGain(BasePrimitive):
    """
    Convert raw data numbers to electrons.

    Uses the ATSECn FITS header keywords to divide image into amp regions and
    then corrects each region with the corresponding GAINn keyword.  Updates the
    following FITS header keywords:

        * GAINCOR: sets to ``True`` if operation performed.
        * BUNIT: sets to `electron`.
        * HISTORY: records the operation.

    Uses the following configuration parameter:

        * saveintims: if set to ``True`` write out a \*_gain.fits file with gain corrected.  Default is ``False``.

    Updates image in returned arguments.

    """

    def __init__(self, action, context):
        BasePrimitive.__init__(self, action, context)
        self.logger = context.pipeline_logger

    def _perform(self):
        # Header keyword to update
        key = 'GAINCOR'
        keycom = 'Gain corrected?'
        # print(self.action.args.ccddata.header)
        number_of_amplifiers = self.action.args.namps
        bsec, dsec, tsec, direc, amps, aoff = self.action.args.map_ccd
        namps = len(amps)
        if namps != number_of_amplifiers:
            self.logger.warning("Amp count disagreement!")
        for amplifier in amps:
            # get amp section
            section = self.action.args.ccddata.header['ATSEC%d' %
                                                      amplifier]
            parsed_section, read_forward = parse_imsec(section)
            # get gain for this amp
            gain = self.action.args.ccddata.header['GAIN%d' % amplifier]
            self.logger.info(
                "Applying gain correction of %.3f in section %s" %
                (gain, self.action.args.ccddata.header['ATSEC%d' %
                                                       amplifier]))
            self.action.args.ccddata.data[
                parsed_section[0]:(parsed_section[1]+1),
                parsed_section[2]:(parsed_section[3]+1)] *= gain

        self.action.args.ccddata.header[key] = (True, keycom)
        self.action.args.ccddata.header['BUNIT'] = ('electron', 'Pixel units')
        self.action.args.ccddata.unit = 'electron'

        log_string = CorrectGain.__module__
        self.action.args.ccddata.header['HISTORY'] = log_string
        self.logger.info(log_string)

        if self.config.instrument.saveintims:
            kcwi_fits_writer(self.action.args.ccddata,
                             table=self.action.args.table,
                             output_file=self.action.args.name,
                             output_dir=self.config.instrument.output_directory,
                             suffix="gain")
        return self.action.args
    # END: class CorrectGain()



from keckdrpframework.primitives.base_primitive import BasePrimitive
from kcwidrp.primitives.kcwi_file_primitives import kcwi_fits_writer

import numpy as np


class TrimOverscan(BasePrimitive):
    """
    Trim overscan region from image.

    Uses the data section (DSECn header keyword) to determine how to trim the
    image to exclude the overscan region.

    Removes raw section keywords, ASECn, BSECn, CSECn, and DSECn after trimming
    and replaces them with the ATSECn keyword giving the image section in the
    trimmed image for each amplifier.

    Uses the following configuration parameter:

        * saveintims: if set to ``True`` write out the trimmed image as \*_trim.fits.  Default is ``False``.

    If the input image is a bias frames, writes out a \*_intb.fits file,
    otherwise, just updates the image in the returned arguments.

    """

    def __init__(self, action, context):
        BasePrimitive.__init__(self, action, context)
        self.logger = context.pipeline_logger

    def _perform(self):

        # parameters
        # image sections for each amp
        bsec, dsec, tsec, direc, amps, aoff = self.action.args.map_ccd
        # handle biases
        is_bias = ('BIAS' in self.action.args.ccddata.header['IMTYPE'])
        # header keyword to update
        key = 'OSCANTRM'
        keycom = 'Overscan trimmed?'
        # get output image dimensions
        max_sec = max(tsec)
        # do we have flags?
        have_flags = (self.action.args.ccddata.flags is not None)
        # create new blank image
        new = np.zeros((max_sec[1]+1, max_sec[3]+1), dtype=np.float32)
        if have_flags:
            new_flags = np.zeros((max_sec[1]+1, max_sec[3]+1), dtype=np.uint8)
        else:
            new_flags = None
        # loop over amps
        for ia in amps:
            # bias correct amp number for indexing python arrays
            iac = ia - aoff
            # input range indices
            yi0 = dsec[iac][0]
            yi1 = dsec[iac][1] + 1
            xi0 = dsec[iac][2]
            xi1 = dsec[iac][3] + 1
            # output range indices
            yo0 = tsec[iac][0]
            yo1 = tsec[iac][1] + 1
            xo0 = tsec[iac][2]
            xo1 = tsec[iac][3] + 1
            # transfer to new image
            new[yo0:yo1, xo0:xo1] = self.action.args.ccddata.data[yi0:yi1,
                                                                  xi0:xi1]
            if have_flags:
                new_flags[yo0:yo1,
                          xo0:xo1] = self.action.args.ccddata.flags[yi0:yi1,
                                                                    xi0:xi1]
            # update amp section
            sec = "[%d:" % (xo0+1)
            sec += "%d," % xo1
            sec += "%d:" % (yo0+1)
            sec += "%d]" % yo1
            self.action.args.ccddata.header[
                'ATSEC%d' % ia] = (sec, "Amp section in trimmed image")
            # remove obsolete sections
            self.action.args.ccddata.header.pop('ASEC%d' % ia)
            self.action.args.ccddata.header.pop('BSEC%d' % ia)
            self.action.args.ccddata.header.pop('DSEC%d' % ia)
            self.action.args.ccddata.header.pop('CSEC%d' % ia)
            # only in RED images
            t_key = 'TSEC%d' % ia
            if t_key in self.action.args.ccddata.header:
                self.action.args.ccddata.header.pop(t_key)
        # update with new image
        self.action.args.ccddata.data = new
        self.action.args.ccddata.flags = new_flags
        self.action.args.ccddata.header['NAXIS1'] = max_sec[3] + 1
        self.action.args.ccddata.header['NAXIS2'] = max_sec[1] + 1
        self.action.args.ccddata.header[key] = (True, keycom)

        log_string = TrimOverscan.__module__
        self.action.args.ccddata.header['HISTORY'] = log_string
        self.logger.info(log_string)

        if self.config.instrument.saveintims:
            kcwi_fits_writer(
                self.action.args.ccddata, table=self.action.args.table,
                output_file=self.action.args.name,
                output_dir=self.config.instrument.output_directory,
                suffix="trim")
        if is_bias:
            kcwi_fits_writer(
                self.action.args.ccddata, table=self.action.args.table,
                output_file=self.action.args.name,
                output_dir=self.config.instrument.output_directory,
                suffix="intb")
            self.context.proctab.update_proctab(
                frame=self.action.args.ccddata, suffix="intb", newtype='BIAS',
                filename=self.action.args.ccddata.header['OFNAME'])
            self.context.proctab.write_proctab(
                tfil=self.config.instrument.procfile)
        return self.action.args
    # END: class TrimOverscan()



from keckdrpframework.primitives.base_primitive import BasePrimitive
from kcwidrp.primitives.kcwi_file_primitives import kcwi_fits_reader, \
        get_master_name
import os


class SubtractDark(BasePrimitive):
    """
    Subtract the master dark frame.

    Checks for existence of master dark frame and, if it exists, performs
    subtraction and records processing in the header.

    """

    def __init__(self, action, context):
        BasePrimitive.__init__(self, action, context)
        self.logger = context.pipeline_logger

    def _perform(self):

        # Header keyword to update
        key = 'DARKSUB'
        keycom = 'master dark subtracted?'
        target_type = 'MDARK'

        self.logger.info("Subtracting master dark")
        tab = self.context.proctab.search_proctab(
            frame=self.action.args.ccddata, target_type=target_type,
            nearest=True)
        self.logger.info("%d master dark frames found" % len(tab))

        if len(tab) > 0:
            mdname = get_master_name(tab, target_type)
            print("*************** READING IMAGE: %s" % mdname)
            mdark = kcwi_fits_reader(
                os.path.join(self.config.instrument.cwd, 'redux',
                             mdname))[0]
            # scale by exposure time
            fac = 1.0
            if 'TTIME' in mdark.header and \
               'TTIME' in self.action.args.ccddata.header:
                fac = float(self.action.args.ccddata.header['TTIME']) / \
                      float(mdark.header['TTIME'])
                self.logger.info("dark scaled by %.3f" % fac)
            else:
                self.logger.warn("unable to scale dark by exposure time")

            # do the subtraction
            self.action.args.ccddata.data -= mdark.data * fac

            self.action.args.ccddata.header[key] = (True, keycom)
            self.action.args.ccddata.header['MDFILE'] = (mdname,
                                                         "Master dark filename")
            self.action.args.ccddata.header['DARKSCL'] = (fac,
                                                          "dark scale factor")
        else:
            self.logger.info("No master dark frame available, skipping")
            self.action.args.ccddata.header[key] = (False, keycom)

        log_string = SubtractDark.__module__
        self.action.args.ccddata.header['HISTORY'] = log_string
        self.logger.info(log_string)

        return self.action.args
    # END: class SubtractDark()



from keckdrpframework.primitives.base_primitive import BasePrimitive
from kcwidrp.primitives.kcwi_file_primitives import kcwi_fits_reader, \
    kcwi_fits_writer, strip_fname  # , get_master_name
import os


class SubtractSky(BasePrimitive):
    """
    Subtract sky model from observation.

    Reads in model of sky created in MakeMasterSky.py and subtracts it from
    observation.  Which model is determined by the `kcwi.sky` file, if present.
    If not, then the model generated from the observation itself will be used.

    Which master sky was used will be recorded in the header keyword SKYMAST.
    If a mask file was used to generate the sky model, it will be recorded in
    SKYMSKF.  If scaling was required prior to subtraction due to differing
    exposure times of observation and model, the scale value is recorded in
    SKYSCL.

    Writes out a \*_intk.fits file and adds an entry to the proc file.
    """

    def __init__(self, action, context):
        BasePrimitive.__init__(self, action, context)
        self.logger = context.pipeline_logger

    def _pre_condition(self):
        """
        Checks if a master sky exists to subtract
        """
        self.logger.info("Checking precondition for SubtractSky")
        keycom = 'sky corrected?'

        if self.config.instrument.skipsky:
            self.logger.warning("Sky subtraction turned off, "
                                "skipping SubtractSky")
            self.action.args.ccddata.header['SKYCOR'] = (False, keycom)
            return False

        skyfile = None
        skymask = None
        # check if kcwi.sky exists
        if os.path.exists('kcwi.sky'):
            f = open('kcwi.sky')
            skyproc = f.readlines()
            f.close()
            # is our file in the list?
            ofn = self.action.args.name
            for row in skyproc:
                # skip comments
                if row.startswith('#'):
                    continue
                # skip empty lines
                if len(row.split()) < 1:
                    continue
                # Parse row:
                # <raw sci file> <raw sky file> <optional mask file>
                #  OR
                # <raw sci file> skip
                # to disable sky subtraction
                # Find match to current file
                if ofn in row.split()[0]:
                    skyfile = row.split()[1]
                    # Should we skip sky subtraction?
                    if 'skip' in skyfile:
                        self.logger.info("Skipping sky subtraction for %s" %
                                         ofn)
                        self.action.args.ccddata.header['SKYCOR'] = (False,
                                                                     keycom)
                        return False

                    elif 'cont' in skyfile:
                        self.logger.info("Using continuum source local sky")

                    # Do we have an optional sky mask file?
                    elif len(row.split()) > 2:
                        skymask = row.split()[2]
                        self.logger.info("Found sky mask entry for %s: %s"
                                         % (ofn, skymask))

                    self.logger.info("Found sky entry for %s: %s" % (ofn,
                                                                     skyfile))
            # check if requested files exist
            if skyfile:
                if not os.path.exists(skyfile):
                    skyfile = None
            if skymask:
                if not os.path.exists(skymask):
                    skymask = None
        self.action.args.skyfile = skyfile
        self.action.args.skymask = skymask
        if skyfile:
            self.logger.info("pre condition got 1 master sky, expected 1")
            return True
        else:
            target_type = 'SKY'
            tab = self.context.proctab.search_proctab(
                frame=self.action.args.ccddata, target_type=target_type,
                nearest=True)
            self.logger.info("pre condition got %d master sky, expected 1"
                             % len(tab))
            if len(tab) <= 0:
                return False
            else:
                return True

    def _perform(self):
        self.logger.info("Subtracting sky background")

        # Header keyword to update
        key = 'SKYCOR'
        keycom = 'sky corrected?'
        target_type = 'SKY'

        skyfile = self.action.args.skyfile
        skymask = self.action.args.skymask

        if not self.action.args.skyfile:
            tab = self.context.proctab.search_proctab(
                frame=self.action.args.ccddata, target_type=target_type,
                nearest=True)
            self.logger.info("%d master sky frames found" % len(tab))

            if len(tab) > 0:
                skyfile = tab['filename'][0]

        msname = strip_fname(skyfile) + '_' + target_type.lower() + ".fits"
        if os.path.exists(os.path.join(self.config.instrument.cwd,
                                       'redux', msname)):
            self.logger.info("Reading image: %s" % msname)
            msky = kcwi_fits_reader(
                os.path.join(self.config.instrument.cwd, 'redux',
                             msname))[0]

            # scale the sky?
            obtime = self.action.args.ccddata.header['XPOSURE']
            sktime = msky.header['XPOSURE']

            if obtime <= 0. or sktime <= 0.:
                self.logger.warning("Bad exposure times (obj, sky): %.1f, %1f"
                                    % (obtime, sktime))
                skscl = 1.
            else:
                skscl = obtime / sktime
            self.logger.info("Sky scale factor = %.3f" % skscl)

            # store un-sky-subtracted image
            self.action.args.ccddata.noskysub = \
                self.action.args.ccddata.data.copy()

            # do the subtraction
            self.action.args.ccddata.data -= msky.data * skscl

            # update header keywords
            self.action.args.ccddata.header[key] = (True, keycom)
            self.action.args.ccddata.header['SKYMAST'] = (msname,
                                                          "Master sky filename")
            self.action.args.ccddata.header['SKYSCL'] = (skscl,
                                                         'sky scale factor')
            if skymask:
                self.action.args.ccddata.header['SKYMSKF'] = (skymask,
                                                              'sky mask file')

        else:
            # update header keywords
            self.action.args.ccddata.header[key] = (False, keycom)

        log_string = SubtractSky.__module__
        self.action.args.ccddata.header['HISTORY'] = log_string

        # write out int image
        kcwi_fits_writer(self.action.args.ccddata,
                         table=self.action.args.table,
                         output_file=self.action.args.name,
                         output_dir=self.config.instrument.output_directory,
                         suffix="intk")
        self.context.proctab.update_proctab(frame=self.action.args.ccddata,
                                            suffix="intk",
                                            filename=self.action.args.name)
        self.context.proctab.write_proctab(tfil=self.config.instrument.procfile)

        self.logger.info(log_string)

        return self.action.args
    # END: class SubtractSky()



from keckdrpframework.primitives.base_primitive import BasePrimitive
from kcwidrp.primitives.kcwi_file_primitives import parse_imsec

from astropy.nddata import StdDevUncertainty

import numpy as np


class CreateUncertaintyImage(BasePrimitive):
    """
    Generate a standard deviation uncertainty image.

    Starts with pure Poisson noise and uses astropy.nddata.StdDevUncertainty to
    generate the uncertainty frame.  If BIASRNn header keywords present, uses
    these with the ATSECn keywords to apply the appropriate readnoise addition
    to each amplifier region.

    """

    def __init__(self, action, context):
        BasePrimitive.__init__(self, action, context)
        self.logger = context.pipeline_logger

    def _perform(self):
        """Assumes units of image are electron"""

        # Header keyword to update
        key = 'UNCSTD'
        keycom = 'stddev uncertainty created?'

        self.logger.info("Create uncertainty image")
        # start with Poisson noise
        self.action.args.ccddata.uncertainty = StdDevUncertainty(
            np.sqrt(np.abs(self.action.args.ccddata.data)), copy=True)
        # add readnoise, if known
        have_bias = False
        bsec, dsec, tsec, direc, amps, aoff = self.action.args.map_ccd
        for amp in amps:
            if 'BIASRN%d' % amp in self.action.args.ccddata.header:
                have_bias = True
        if have_bias:
            number_of_amplifiers = self.action.args.ccddata.header['NVIDINP']
            namps = len(amps)
            if namps != number_of_amplifiers:
                self.logger.warning("Amp count disagreement!")
            for amplifier in amps:
                # get amp parameters
                bias_readnoise = self.action.args.ccddata.header[
                    'BIASRN%d' % amplifier]
                section = self.action.args.ccddata.header['ATSEC%d' %
                                                          amplifier]
                parsed_section, read_forward = parse_imsec(section)
                self.action.args.ccddata.uncertainty.array[
                    parsed_section[0]:(parsed_section[1]+1),
                    parsed_section[2]:(parsed_section[3]+1)] = \
                    np.sqrt(
                    self.action.args.ccddata.uncertainty.array[
                        parsed_section[0]:(parsed_section[1] + 1),
                        parsed_section[2]:(parsed_section[3] + 1)] ** 2 +
                        bias_readnoise ** 2)
        else:
            self.logger.warn("Readnoise undefined, uncertainty Poisson only")
        # check for flags and mask arrays
        if self.action.args.ccddata.flags is None:
            self.action.args.ccddata.flags = np.zeros(
                self.action.args.ccddata.data.shape, dtype=np.uint8)
        if self.action.args.ccddata.mask is None:
            self.action.args.ccddata.mask = np.zeros(
                self.action.args.ccddata.data.shape, dtype=np.uint8)
        # document variance image creation
        self.action.args.ccddata.header[key] = (True, keycom)

        log_string = CreateUncertaintyImage.__module__
        self.action.args.ccddata.header['HISTORY'] = log_string
        self.logger.info(log_string)

        return self.action.args
    # END: class CreateUncertaintyImage()



from keckdrpframework.primitives.base_primitive import BasePrimitive
from kcwidrp.primitives.kcwi_file_primitives import kcwi_fits_writer

import numpy as np
from astropy.nddata import CCDData


class CubeImage(BasePrimitive):
    """
    Transform 3D data cube into 2D image.

    Generates a wavelength-aligned 2d image of a 3d data cube.  This is done
    automatically for arc lamps used for the wavelength solution as a
    diagnostic for the wavelength fit.

    Writes out image with "_icube_2d.fits" suffix.

    """

    def __init__(self, action, context):
        BasePrimitive.__init__(self, action, context)
        self.logger = context.pipeline_logger

    def _pre_condition(self):
        """Checks if we have a cube to convert"""
        self.logger.info("Checking precondition for CubeImage")
        sz = self.action.args.ccddata.data.shape
        if len(sz) == 3:
            if sz[2] == 24:
                self.logger.info("CubeImage precondition met.")
                self.action.args.cube_size = sz
                return True
            else:
                self.logger.warning("Mal-formed cube!")
                return False
        else:
            self.logger.warning("Not a 3D data cube!")
            return False

    def _perform(self):
        self.logger.info("Creating 2D image from 3D cube")

        log_string = CubeImage.__module__

        # get wavelength ranges
        wall0 = self.action.args.ccddata.header['WAVALL0']
        wall1 = self.action.args.ccddata.header['WAVALL1']
        w0 = self.action.args.ccddata.header['CRVAL3']
        dw = self.action.args.ccddata.header['CD3_3']
        crpixw = self.action.args.ccddata.header['CRPIX3']
        y0 = int((wall0 - w0) / dw)
        if y0 < 0:
            y0 = 0
        out_w0 = y0 * dw + w0
        y1 = int((wall1 - w0) / dw)
        if y1 > self.action.args.cube_size[0]:
            y1 = self.action.args.cube_size[0]
        out_y = y1 - y0
        # create output image
        cub_x = self.action.args.cube_size[1]
        out_x = int(24 * cub_x)
        out_img = np.zeros((out_y, out_x), dtype=np.float)
        # set spatial scale
        s0 = 0.
        ds = 24.0 / out_x
        crpixs = 1.
        self.logger.info("cube dims: y, x, z: %d, %d, %d" %
                         self.action.args.cube_size)
        self.logger.info("y0, y1 = %d, %d: out_x, out_y = %d, %d" %
                         (y0, y1, out_x, out_y))

        # loop over slices
        for isl in range(24):
            x0 = isl * cub_x
            x1 = (isl + 1) * cub_x
            out_img[:, x0:x1] = self.action.args.ccddata.data[y0:y1, :, isl]
        # output CCDData structure
        out_ccd = CCDData(out_img, meta=self.action.args.ccddata.header,
                          unit=self.action.args.ccddata.unit)

        # update header
        out_ccd.header['HISTORY'] = log_string
        del out_ccd.header['RADESYS']
        del out_ccd.header['LONPOLE']
        del out_ccd.header['LATPOLE']
        del out_ccd.header['CTYPE1']
        del out_ccd.header['CTYPE2']
        del out_ccd.header['CTYPE3']
        del out_ccd.header['CUNIT1']
        del out_ccd.header['CUNIT2']
        del out_ccd.header['CUNIT3']
        del out_ccd.header['CNAME1']
        del out_ccd.header['CNAME2']
        del out_ccd.header['CNAME3']
        del out_ccd.header['CRVAL1']
        del out_ccd.header['CRVAL2']
        del out_ccd.header['CRVAL3']
        del out_ccd.header['CRPIX1']
        del out_ccd.header['CRPIX2']
        del out_ccd.header['CRPIX3']
        del out_ccd.header['CD1_1']
        del out_ccd.header['CD1_2']
        del out_ccd.header['CD2_1']
        del out_ccd.header['CD2_2']
        del out_ccd.header['CD3_3']

        out_ccd.header['WCSDIM'] = 2

        out_ccd.header['CTYPE1'] = ('SPATIAL', 'slice number')
        out_ccd.header['CUNIT1'] = ('slu', 'slice units')
        out_ccd.header['CNAME1'] = ('KCWI Slice', 'slice name')
        out_ccd.header['CRVAL1'] = (s0, 'slice zeropoint')
        out_ccd.header['CRPIX1'] = (crpixs, 'slice reference pixel')
        out_ccd.header['CDELT1'] = (ds, 'slice units per pixel')

        out_ccd.header['CTYPE2'] = ('AWAV', 'Air Wavelengths')
        out_ccd.header['CUNIT2'] = ('Angstrom', 'Wavelength units')
        out_ccd.header['CNAME2'] = ('KCWI Wavelength', 'Wavelength name')
        out_ccd.header['CRVAL2'] = (out_w0, 'Wavelength zeropoint')
        out_ccd.header['CRPIX2'] = (crpixw, 'Wavelength reference pixel')
        out_ccd.header['CDELT2'] = (dw, 'Wavelength Angstroms per pixel')

        # write out image
        kcwi_fits_writer(out_ccd, output_file=self.action.args.name,
                         output_dir=self.config.instrument.output_directory,
                         suffix="icube_2d")

        self.logger.info(log_string)

        return self.action.args
    # END: class CubeImage()



from keckdrpframework.primitives.base_primitive import BasePrimitive
from kcwidrp.core.kcwi_get_std import kcwi_get_std
from kcwidrp.primitives.kcwi_file_primitives import kcwi_fits_writer, \
    kcwi_fits_reader, strip_fname

import numpy as np
from scipy.ndimage import shift
from astropy.nddata import CCDData
import math
import ref_index
import os


def atm_disper(w0, w1, airmass, temperature=10.0, pressure_pa=61100.0,
               humidity=50.0, co2=400.0):
    """

    Calculate atmospheric dispersion at w1 relative to w0

    Args:
        w0 (float): reference wavelength (Angstroms)
        w1 (float): offset wavelength (Angstroms)
        airmass (float): unitless airmass
        temperature (float): atmospheric temperature (C)
        pressure_pa (float): atmospheric pressure (Pa)
        humidity (float): relative humidity (%)
        co2 (float): Carbon-Dioxide (mu-mole/mole)

    """

    # Calculate
    z = math.acos(1.0/airmass)

    n0 = ref_index.ciddor(wave=w0/10., t=temperature, p=pressure_pa,
                          rh=humidity, co2=co2)
    n1 = ref_index.ciddor(wave=w1/10., t=temperature, p=pressure_pa,
                          rh=humidity, co2=co2)

    return 206265.0 * (n0 - n1) * math.tan(z)


class CorrectDar(BasePrimitive):
    """
    Correct for Differential Atmospheric Refraction

    Accounts for rotator orientation, zenith angle, and parallactic angle to
    correct input data cube into a padded, DAR corrected output cube.
    Calculates the DAR correction for each wavelength slice and adjusts position
    in cube using scipy.nddata.shift to implement correction.

    Sets flags in padded region of output cube to a value of 128.

    Also corrects delta wavelength cube if it is a standard star observation.

    """

    def __init__(self, action, context):
        BasePrimitive.__init__(self, action, context)
        self.logger = context.pipeline_logger

    def _pre_condition(self):
        """Checks if DAR correction is appropriate"""
        self.logger.info("Checking precondition for CorrectDar")
        # Check image
        if 'GEOMCOR' not in self.action.args.ccddata.header:
            self.logger.error("Can only correct DAR on geometrically corrected "
                              "images")
            self.action.args.ccddata.header['DARCOR'] = (False,
                                                         'DAR corrected?')
            return False
        else:
            if not self.action.args.ccddata.header['GEOMCOR']:
                self.logger.error(
                    "Can only correct DAR on geometrically corrected "
                    "images")
                self.action.args.ccddata.header['DARCOR'] = (False,
                                                             'DAR corrected?')
                return False
            else:
                return True

    def _perform(self):
        """Correct for differential atmospheric refraction"""
        self.logger.info("Correcting for DAR")
        # image size
        image_size = self.action.args.ccddata.data.shape

        # get wavelengths
        w0 = self.action.args.ccddata.header['CRVAL3']
        dw = self.action.args.ccddata.header['CD3_3']
        waves = w0 + np.arange(image_size[0]) * dw
        wgoo0 = self.action.args.ccddata.header['WAVGOOD0']
        wgoo1 = self.action.args.ccddata.header['WAVGOOD1']
        wref = self.action.args.ccddata.header['WAVMID']
        self.logger.info("Ref WL = %.1f, good WL range = (%.1f - %.1f)" %
                         (wref, wgoo0, wgoo1))

        # spatial scales in arcsec/item
        y_scale = self.action.args.ccddata.header['PXSCL'] * 3600.
        x_scale = self.action.args.ccddata.header['SLSCL'] * 3600.

        # padding depends on grating
        if 'H' in self.action.args.grating:
            padding_as = 2.0
        elif 'M' in self.action.args.grating:
            padding_as = 3.0
        else:
            padding_as = 4.0

        padding_x = int(padding_as / x_scale)
        padding_y = int(padding_as / y_scale)

        # update WCS
        crpix1 = self.action.args.ccddata.header['CRPIX1']
        crpix2 = self.action.args.ccddata.header['CRPIX2']
        self.action.args.ccddata.header['CRPIX1'] = crpix1 + float(padding_x)
        self.action.args.ccddata.header['CRPIX2'] = crpix2 + float(padding_y)

        # airmass
        airmass = self.action.args.ccddata.header['AIRMASS']
        self.logger.info("Airmass: %.3f" % airmass)

        # IFU orientation
        ifu_pa = self.action.args.ccddata.header['IFUPA']

        # Parallactic angle
        parallactic_angle = self.action.args.ccddata.header['PARANG']

        # Projection angle in radians
        projection_angle_deg = ifu_pa - parallactic_angle
        projection_angle = math.radians(projection_angle_deg)
        self.logger.info("DAR Angles: ifu_pa, parang, projang (deg): "
                         "%.2f, %.2f, %.2f" % (ifu_pa, parallactic_angle,
                                               projection_angle_deg))

        # dispersion over goo wl range in arcsec
        dispersion_max_as = atm_disper(wgoo1, wgoo0, airmass)

        # projected onto IFU
        xdmax_as = dispersion_max_as * math.sin(projection_angle)
        ydmax_as = dispersion_max_as * math.cos(projection_angle)
        self.logger.info("DAR over GOOD WL range: total, x, y (asec): "
                         "%.2f, %.2f, %.2f" % (dispersion_max_as, xdmax_as,
                                               ydmax_as))

        # now in pixels
        xdmax_px = xdmax_as / x_scale
        ydmax_px = ydmax_as / y_scale
        dmax_px = math.sqrt(xdmax_px**2 + ydmax_px**2)
        self.logger.info("DAR over GOOD WL range: total, x, y (pix): "
                         "%.2f, %.2f, %.2f" % (dmax_px, xdmax_px, ydmax_px))

        # prepare output cubes
        output_image = np.zeros((image_size[0], image_size[1]+2*padding_y,
                                 image_size[2]+2*padding_x), dtype=np.float64)
        output_stddev = output_image.copy()
        output_mask = np.zeros((image_size[0], image_size[1]+2*padding_y,
                                image_size[2]+2*padding_x), dtype=np.uint8)
        output_flags = np.zeros((image_size[0], image_size[1] + 2*padding_y,
                                 image_size[2] + 2 * padding_x), dtype=np.uint8)
        if self.action.args.ccddata.noskysub is not None:
            output_noskysub = np.zeros((image_size[0],
                                        image_size[1] + 2*padding_y,
                                        image_size[2] + 2 * padding_x),
                                       dtype=np.float64)
        else:
            output_noskysub = None
        # DAR padded pixel flag
        output_flags += 128

        output_image[:, padding_y:(padding_y+image_size[1]),
                     padding_x:(padding_x+image_size[2])] = \
            self.action.args.ccddata.data

        output_stddev[:, padding_y:(padding_y+image_size[1]),
                      padding_x:(padding_x+image_size[2])] = \
            self.action.args.ccddata.uncertainty.array

        output_mask[:, padding_y:(padding_y+image_size[1]),
                    padding_x:(padding_x+image_size[2])] = \
            self.action.args.ccddata.mask

        output_flags[:, padding_y:(padding_y+image_size[1]),
                     padding_x:(padding_x+image_size[2])] = \
            self.action.args.ccddata.flags

        if output_noskysub is not None:
            output_noskysub[:, padding_y:(padding_y + image_size[1]),
                            padding_x:(padding_x + image_size[2])] = \
                self.action.args.ccddata.noskysub

        # check for obj, sky cubes
        output_obj = None
        output_sky = None
        if self.action.args.nasmask and self.action.args.numopen > 1:
            ofn = self.action.args.name

            objfn = strip_fname(ofn) + '_ocube.fits'
            full_path = os.path.join(
                self.config.instrument.cwd,
                self.config.instrument.output_directory, objfn)
            if os.path.exists(full_path):
                obj = kcwi_fits_reader(full_path)[0]
                output_obj = np.zeros(
                    (image_size[0], image_size[1] + 2 * padding_y,
                     image_size[2] + 2 * padding_x), dtype=np.float64)
                output_obj[:, padding_y:(padding_y + image_size[1]),
                           padding_x:(padding_x + image_size[2])] = obj.data

            skyfn = strip_fname(ofn) + '_scube.fits'
            full_path = os.path.join(
                self.config.instrument.cwd,
                self.config.instrument.output_directory, skyfn)
            if os.path.exists(full_path):
                sky = kcwi_fits_reader(full_path)[0]
                output_sky = np.zeros(
                    (image_size[0], image_size[1] + 2 * padding_y,
                     image_size[2] + 2 * padding_x), dtype=np.float64)
                output_sky[:, padding_y:(padding_y + image_size[1]),
                           padding_x:(padding_x + image_size[2])] = sky.data

        # check if we have a standard star observation
        output_del = None
        stdfile, _ = kcwi_get_std(self.action.args.ccddata.header['OBJECT'],
                                  self.logger)
        if stdfile is not None:
            afn = self.action.args.ccddata.header['ARCFL']

            delfn = strip_fname(afn) + '_dcube.fits'
            full_path = os.path.join(
                self.config.instrument.cwd,
                self.config.instrument.output_directory, delfn)
            if os.path.exists(full_path):
                dew = kcwi_fits_reader(full_path)[0]
                output_del = np.zeros(
                    (image_size[0], image_size[1] + 2 * padding_y,
                     image_size[2] + 2 * padding_x), dtype=np.float64)
                output_del[:, padding_y:(padding_y + image_size[1]),
                           padding_x:(padding_x + image_size[2])] = dew.data

        # Perform correction
        for j, wl in enumerate(waves):
            dispersion_correction = atm_disper(wref, wl, airmass)
            x_shift = dispersion_correction * \
                math.sin(projection_angle) / x_scale
            y_shift = dispersion_correction * \
                math.cos(projection_angle) / y_scale
            output_image[j, :, :] = shift(output_image[j, :, :],
                                          (y_shift, x_shift))
            output_stddev[j, :, :] = shift(output_stddev[j, :, :],
                                           (y_shift, x_shift))
            output_mask[j, :, :] = shift(output_mask[j, :, :],
                                         (y_shift, x_shift))
            output_flags[j, :, :] = shift(output_flags[j, :, :],
                                          (y_shift, x_shift))
            if output_noskysub is not None:
                output_noskysub[j, :, :] = shift(output_noskysub[j, :, :],
                                                 (y_shift, x_shift))
        # for obj, sky if they exist
        if output_obj is not None:
            for j, wl in enumerate(waves):
                dispersion_correction = atm_disper(wref, wl, airmass)
                x_shift = dispersion_correction * \
                    math.sin(projection_angle) / x_scale
                y_shift = dispersion_correction * \
                    math.cos(projection_angle) / y_scale
                output_obj[j, :, :] = shift(output_obj[j, :, :],
                                            (y_shift, x_shift))

        if output_sky is not None:
            for j, wl in enumerate(waves):
                dispersion_correction = atm_disper(wref, wl, airmass)
                x_shift = dispersion_correction * \
                    math.sin(projection_angle) / x_scale
                y_shift = dispersion_correction * \
                    math.cos(projection_angle) / y_scale
                output_sky[j, :, :] = shift(output_sky[j, :, :],
                                            (y_shift, x_shift))

        # for delta wavelength cube, if it exists
        if output_del is not None:
            for j, wl in enumerate(waves):
                dispersion_correction = atm_disper(wref, wl, airmass)
                x_shift = dispersion_correction * \
                    math.sin(projection_angle) / x_scale
                y_shift = dispersion_correction * \
                    math.cos(projection_angle) / y_scale
                output_del[j, :, :] = shift(output_del[j, :, :],
                                            (y_shift, x_shift))

        self.action.args.ccddata.data = output_image
        self.action.args.ccddata.uncertainty.array = output_stddev
        self.action.args.ccddata.mask = output_mask
        self.action.args.ccddata.flags = output_flags
        self.action.args.ccddata.noskysub = output_noskysub

        log_string = CorrectDar.__module__

        # update header
        self.action.args.ccddata.header['HISTORY'] = log_string
        self.action.args.ccddata.header['DARCOR'] = (True, 'DAR corrected?')
        self.action.args.ccddata.header['DARANG'] = (projection_angle_deg,
                                                     'DAR projection angle '
                                                     '(deg)')
        self.action.args.ccddata.header['DARPADX'] = (padding_x,
                                                      'DAR X padding (pix)')
        self.action.args.ccddata.header['DARPADY'] = (padding_y,
                                                      'DAR Y padding (pix)')
        self.action.args.ccddata.header['DAREFWL'] = (wref,
                                                      'DAR reference wl (Ang)')
        # write out corrected image
        kcwi_fits_writer(self.action.args.ccddata,
                         table=self.action.args.table,
                         output_file=self.action.args.name,
                         output_dir=self.config.instrument.output_directory,
                         suffix="icubed")
        self.context.proctab.update_proctab(frame=self.action.args.ccddata,
                                            suffix="icubed",
                                            filename=self.action.args.name)
        self.context.proctab.write_proctab(tfil=self.config.instrument.procfile)

        # check for sky, obj cube
        if output_obj is not None:
            out_obj = CCDData(output_obj,
                              meta=self.action.args.ccddata.header,
                              unit=self.action.args.ccddata.unit)
            kcwi_fits_writer(
                out_obj, output_file=self.action.args.name,
                output_dir=self.config.instrument.output_directory,
                suffix="ocubed")
        if output_sky is not None:
            out_sky = CCDData(output_sky,
                              meta=self.action.args.ccddata.header,
                              unit=self.action.args.ccddata.unit)
            kcwi_fits_writer(
                out_sky, output_file=self.action.args.name,
                output_dir=self.config.instrument.output_directory,
                suffix="scubed")

        # check for delta wave cube
        if output_del is not None:
            out_del = CCDData(output_del,
                              meta=self.action.args.ccddata.header,
                              unit=self.action.args.ccddata.unit)
            kcwi_fits_writer(
                out_del, output_file=self.action.args.name,
                output_dir=self.config.instrument.output_directory,
                suffix="dcubed")

        self.logger.info(log_string)

        return self.action.args
    # END: class CorrectDar()



from keckdrpframework.primitives.base_img import BaseImg
from kcwidrp.primitives.kcwi_file_primitives import kcwi_fits_reader, \
    kcwi_fits_writer, strip_fname
from kcwidrp.primitives.GetAtlasLines import gaus
from kcwidrp.core.kcwi_get_std import kcwi_get_std
from kcwidrp.core.bokeh_plotting import bokeh_plot
from kcwidrp.core.kcwi_plotting import save_plot
from kcwidrp.core.bspline import Bspline
from bokeh.plotting import figure

import os
import time
import numpy as np
from scipy.optimize import curve_fit
from astropy.io import fits


class MakeMasterSky(BaseImg):
    """
    Make master sky image.

    Uses b-spline fits along with geometry maps to generate a master sky image
    for sky subtraction.

    This routine also handles the file `kcwi.sky`, which controls the master
    sky generation.  This file consists of one line per image, with the first
    column indicating the raw object image to be sky-subtracted.  The following
    columns can either indicate a separate image to use for sky subtraction, the
    filename of a mask fits image for masking object flux, or indicate that the
    object is a continuum source and either automatically find the object, or
    specify the location and width of the continuum source.  Below are example
    one-line entries and what they mean:

    1. Skip sky subtraction for this particular object image:

        * kr230925_00075.fits skip

    2. Point to a different image for the sky (this assumes the \*_sky.fits
    image has already been generated previously:

        * kr230925_00075.fits kr230925_00076.fits

    3. Indicate that a mask file should be used to mask object flux when
    deriving the sky model (see kcwi_masksky_ds9.py):

        * kr230925_00075.fits kr230925_00075_smsk.fits

    4. Indicate that this is a bright continuum source and automatically mask
    the continuum source from the sky model:

        * kr230925_00075.fits cont

    5. Indicate that this is a faint continuum source and specify the location
    and the width of the continuum source (in pixels):

        * kr230925_00075.fits cont 45.0 7.6

    If no `kcwi.sky` file exists, or there is no entry for the input object
    frame, then the entire image is used to generate the sky model.

    It is good practice to run all the data through first, then inspect the
    sky subtraction and see which frames will benefit from masking or from a
    dedicated sky observation.

    If a sky model is generated, the routine will write out a \*_sky.fits image
    and add a sky entry in the proc table.

    """

    def __init__(self, action, context):
        BaseImg.__init__(self, action, context)
        self.logger = context.pipeline_logger

    def _pre_condition(self):
        """
        Checks if we can create a master sky
        """
        self.logger.info("Checking precondition for MakeMasterSky")

        if self.config.instrument.skipsky:
            self.logger.warning("Sky subtraction turned off, "
                                "skipping MakeMasterSky")
            return False

        suffix = 'sky'  # self.action.args.new_type.lower()
        ofn = self.action.args.name
        rdir = self.config.instrument.output_directory

        # Are we a standard star?
        stdfile = None
        stdname = None
        if 'object' in self.action.args.imtype.lower():
            self.logger.info("Checking OBJECT keyword")
            stdfile, stdname = kcwi_get_std(
                self.action.args.ccddata.header['OBJECT'], self.logger)
            if not stdfile:
                self.logger.info("Checking TARGNAME keyword")
                stdfile, stdname = kcwi_get_std(
                    self.action.args.ccddata.header['TARGNAME'], self.logger)
        else:
            self.logger.warning("Not object type: %s" %
                                self.action.args.imtype)
        self.action.args.stdfile = stdfile
        self.action.args.stdname = stdname

        # Is there a kcwi.sky file?
        skyfile = None
        skymask = None
        contsky = False
        cont_source_pos = None
        cont_source_width = None
        # check if kcwi.sky exists
        if os.path.exists('kcwi.sky'):
            self.logger.info("Reading kcwi.sky")
            f = open('kcwi.sky')
            skyproc = f.readlines()
            f.close()
            # is our file in the list?
            for row in skyproc:
                # skip comments
                if row.startswith('#'):
                    continue
                # skip empty lines
                if len(row.split()) < 1:
                    continue
                # Parse row:
                # <raw sci file> <raw sky file> <optional mask file>
                #  OR
                # <raw sci file> skip
                # to disable sky subtraction
                # Find match to current file
                if ofn in row.split()[0]:
                    skyfile = row.split()[1]
                    # Should we skip sky subtraction?
                    if 'skip' in skyfile:
                        self.logger.info("Skipping sky subtraction for %s" %
                                         ofn)
                        keycom = 'sky corrected?'
                        self.action.args.ccddata.header['SKYCOR'] = (False,
                                                                     keycom)
                        return False
                    elif 'cont' in skyfile:
                        self.logger.info("Using continuum source local sky for"
                                         " %s" % ofn)
                        contsky = True
                        if len(row.split()) == 4:
                            cont_source_pos = float(row.split()[2])
                            cont_source_width = float(row.split()[3])
                            self.logger.info("Using input continuum pos of"
                                             "%.2f with width of %.2f" %
                                             (cont_source_pos,
                                              cont_source_width))

                    # Do we have an optional sky mask file?
                    elif len(row.split()) > 2:
                        skymask = row.split()[2]
                        self.logger.info("Found sky mask entry for %s: %s"
                                         % (ofn, skymask))

                    self.logger.info("Found sky entry for %s: %s" % (ofn,
                                                                     skyfile))
            # Do have a mask file?
            if skymask:
                # Does it exist?
                if os.path.exists(skymask):
                    self.logger.info("Using sky mask file: %s" % skymask)
                else:
                    self.logger.warning("Sky mask file not found: %s" % skymask)
                    skymask = None
        # Record results
        self.action.args.skyfile = skyfile
        self.action.args.skymask = skymask
        self.action.args.contsky = contsky
        self.action.args.cont_source_pos = cont_source_pos
        self.action.args.cont_source_width = cont_source_width
        # Do we have a sky alternate?
        if skyfile:
            # Generate sky file name
            msname = skyfile.split('.fits')[0] + '_' + suffix + '.fits'
            mskyf = os.path.join(rdir, msname)
            # Does it exist?
            if os.path.exists(mskyf):
                self.logger.info("Master sky already exists: %s" % mskyf)
                return False
            else:
                self.logger.warning("Alternate master sky %s not found."
                                    % mskyf)
                return True
        else:
            self.logger.info("No alternate master sky requested.")
            return True

    def _perform(self):
        """
        Returns an Argument() with the parameters that depends on this operation
        """
        self.logger.info("Creating master sky")

        suffix = 'sky'

        # get root for maps
        tab = self.context.proctab.search_proctab(
            frame=self.action.args.ccddata, target_type='MARC',
            target_group=self.action.args.groupid)
        if len(tab) <= 0:
            self.logger.error("Geometry not solved!")
            return self.action.args

        groot = strip_fname(tab['filename'][0])

        # Wavelength map image
        wmf = groot + '_wavemap.fits'
        self.logger.info("Reading image: %s" % wmf)
        wavemap = kcwi_fits_reader(os.path.join(
            self.config.instrument.cwd, 'redux', wmf))[0]

        # Slice map image
        slf = groot + '_slicemap.fits'
        self.logger.info("Reading image: %s" % slf)
        slicemap = kcwi_fits_reader(os.path.join(
            self.config.instrument.cwd, 'redux', slf))[0]

        # Position map image
        pof = groot + '_posmap.fits'
        self.logger.info("Reading image: %s" % pof)
        posmap = kcwi_fits_reader(os.path.join(
            self.config.instrument.cwd, 'redux', pof))[0]
        posmax = np.nanmax(posmap.data)
        posbuf = int(10. / self.action.args.xbinsize)

        ny = posmap.data.shape[0]

        if self.action.args.camera == 1:    # Red
            # Get ymap for trimming junk at ends
            ymap = posmap.copy()
            for i in range(ny):
                ymap.data[i, :] = float(i)
        else:
            ymap = None

        # wavelength region
        wavegood0 = wavemap.header['WAVGOOD0']
        wavegood1 = wavemap.header['WAVGOOD1']
        waveall0 = wavemap.header['WAVALL0']
        waveall1 = wavemap.header['WAVALL1']
        wavemid = wavemap.header['WAVMID']

        # get image size
        sm_sz = self.action.args.ccddata.data.shape

        # sky masking
        # default is no masking (True = mask, False = don't mask)
        binary_mask = np.zeros(sm_sz, dtype=bool)

        # was sky masking requested?
        if self.action.args.skymask:
            if os.path.exists(self.action.args.skymask):
                self.logger.info("Reading sky mask file: %s"
                                 % self.action.args.skymask)
                hdul = fits.open(self.action.args.skymask)
                binary_mask = hdul[0].data
                # verify size match
                bm_sz = binary_mask.shape
                if bm_sz[0] != sm_sz[0] or bm_sz[1] != sm_sz[1]:
                    self.logger.warning("Sky mask size mis-match: "
                                        "masking disabled")
                    binary_mask = np.zeros(sm_sz, dtype=bool)
            else:
                self.logger.warning("Sky mask image not found: %s"
                                    % self.action.args.skymask)

        auto_masked = False
        auto_mask_type = ""
        auto_cont_pos = None
        auto_cont_width = None
        # if we are a standard, get mask for bright continuum source
        if self.action.args.stdname is not None:
            self.logger.info("Standard star observation of "
                             "%s will be auto-masked" %
                             self.action.args.stdname)
            auto_masked = True
            auto_mask_type = "Std Star"
            # Use 10% of wavelength range at wavemid
            std_wav_ran = (wavemid - 0.05 * (wavegood1 - wavegood0),
                           wavemid + 0.05 * (wavegood1 - wavegood0))
            std_sl_max = -1
            std_sl_sig_max = -1.
            std_sl_max_pos_data = None
            std_sl_max_flx_data = None

            self.logger.info("Finding the std max slice")
            for si in range(24):
                sq = [i for i, v in enumerate(slicemap.data.flat) if v == si and
                      std_wav_ran[0] < wavemap.data.flat[i] < std_wav_ran[1] and
                      posbuf < posmap.data.flat[i] < (posmax - posbuf)]
                xplt = posmap.data.flat[sq]
                yplt = self.action.args.ccddata.data.flat[sq]
                sig = float(np.nanstd(yplt))
                self.logger.info("Slice %d - StDev = %.2f" % (si, sig))
                if sig > std_sl_sig_max:
                    std_sl_sig_max = sig
                    std_sl_max = si
                    std_sl_max_pos_data = xplt.copy()
                    std_sl_max_flx_data = yplt.copy()
            ipk = np.argmax(std_sl_max_flx_data)
            ppk = std_sl_max_pos_data[ipk]
            fpk = std_sl_max_flx_data[ipk]

            # gaussian fit to max slice
            res, _ = curve_fit(gaus, std_sl_max_pos_data, std_sl_max_flx_data,
                               p0=[fpk, ppk, 1.])
            self.logger.info("Std max at %.2f in slice %d with width %.2f px"
                             % (res[1], std_sl_max, res[2]))
            std_pos_mask_0 = res[1] - 5. * res[2]
            std_pos_mask_1 = res[1] + 5. * res[2]
            self.logger.info("Masking between %.2f and %.2f" %
                             (std_pos_mask_0, std_pos_mask_1))
            auto_cont_pos = res[1]
            auto_cont_width = 5. * res[2]

            # Mask standard from sky calculation
            for i, v in enumerate(binary_mask.flat):
                if std_pos_mask_0 < posmap.data.flat[i] < std_pos_mask_1:
                    binary_mask.flat[i] = True

            # plot, if requested
            if self.config.instrument.plot_level >= 1:
                xx = np.arange(np.min(std_sl_max_pos_data),
                               np.max(std_sl_max_pos_data), 1)
                yy = gaus(xx, res[0], res[1], res[2])
                p = figure(
                    title=self.action.args.plotlabel +
                    ' Std max sl %d' % std_sl_max,
                    x_axis_label='Pos (x px)',
                    y_axis_label='Flux (e-)',
                    plot_width=self.config.instrument.plot_width,
                    plot_height=self.config.instrument.plot_height)
                p.circle(std_sl_max_pos_data, std_sl_max_flx_data, size=1,
                         line_alpha=0., fill_color='purple',
                         legend_label='Data')
                p.line([ppk, ppk], [0, fpk], color='green')
                p.line([std_pos_mask_0, std_pos_mask_0], [0, fpk], color='blue')
                p.line([std_pos_mask_1, std_pos_mask_1], [0, fpk], color='blue')
                p.line(xx, yy, color='red')
                bokeh_plot(p, self.context.bokeh_session)
                if self.config.instrument.plot_level >= 2:
                    input("Next? <cr>: ")
                else:
                    time.sleep(self.config.instrument.plot_pause)

        # if we are a continuum source,
        # get local mask for bright continuum source
        elif self.action.args.contsky:
            self.logger.info("continuum source observation will be auto-masked")
            auto_masked = True

            # Use 10% of wavelength range at wavemid
            con_wav_ran = (wavemid - 0.05 * (wavegood1 - wavegood0),
                           wavemid + 0.05 * (wavegood1 - wavegood0))
            con_sl_max = -1
            con_sl_sig_max = -1.
            con_sl_max_pos_data = None
            con_sl_max_flx_data = None

            if self.action.args.cont_source_pos is None:
                self.logger.info("Finding the continuum source automatically")
                auto_mask_type = "AutoCont"

                for si in range(24):
                    sq = [i for i, v in enumerate(slicemap.data.flat) if
                          v == si and
                          con_wav_ran[0] < wavemap.data.flat[i] <
                          con_wav_ran[1] and
                          posbuf < posmap.data.flat[i] < (posmax - posbuf)]
                    xplt = posmap.data.flat[sq]
                    yplt = self.action.args.ccddata.data.flat[sq]
                    sig = float(np.nanstd(yplt))
                    self.logger.info("Slice %d - StDev = %.2f" % (si, sig))
                    if sig > con_sl_sig_max:
                        con_sl_sig_max = sig
                        con_sl_max = si
                        con_sl_max_pos_data = xplt.copy()
                        con_sl_max_flx_data = yplt.copy()
                ipk = np.argmax(con_sl_max_flx_data)
                ppk = con_sl_max_pos_data[ipk]
                fpk = con_sl_max_flx_data[ipk]

                # gaussian fit to max slice
                res, _ = curve_fit(gaus, con_sl_max_pos_data,
                                   con_sl_max_flx_data, p0=[fpk, ppk, 1.])
                self.logger.info("Continuum source max at %.2f in "
                                 "slice %d with width %.2f px"
                                 % (res[1], con_sl_max, res[2]))

                # First define source extent
                con_pos_mask_0 = res[1] - 7. * res[2]
                con_pos_mask_1 = res[1] + 7. * res[2]

                auto_cont_pos = res[1]
                auto_cont_width = 7. * res[2]

                # Next define lower and upper windows
                con_pos_mask_lo_0 = con_pos_mask_0 - \
                    14 / self.action.args.xbinsize
                con_pos_mask_up_1 = con_pos_mask_1 + \
                    14 / self.action.args.xbinsize

                # plot, if requested
                if self.config.instrument.plot_level >= 1:
                    xx = np.arange(np.min(con_sl_max_pos_data),
                                   np.max(con_sl_max_pos_data), 1)
                    yy = gaus(xx, res[0], res[1], res[2])
                    p = figure(
                        title=self.action.args.plotlabel +
                        ' Cont source max sl %d' % con_sl_max,
                        x_axis_label='Pos (x px)',
                        y_axis_label='Flux (e-)',
                        plot_width=self.config.instrument.plot_width,
                        plot_height=self.config.instrument.plot_height)
                    p.circle(con_sl_max_pos_data, con_sl_max_flx_data,
                             size=1, line_alpha=0., fill_color='purple',
                             legend_label='Data')
                    p.line([ppk, ppk], [0, fpk], color='green')
                    p.line([con_pos_mask_lo_0, con_pos_mask_lo_0], [0, fpk],
                           color='blue')
                    p.line([con_pos_mask_0, con_pos_mask_0], [0, fpk],
                           color='blue')
                    p.line([con_pos_mask_1, con_pos_mask_1], [0, fpk],
                           color='blue')
                    p.line([con_pos_mask_up_1, con_pos_mask_up_1], [0, fpk],
                           color='blue')
                    p.line(xx, yy, color='red')
                    bokeh_plot(p, self.context.bokeh_session)
                    if self.config.instrument.plot_level >= 2:
                        input("Next? <cr>: ")
                    else:
                        time.sleep(self.config.instrument.plot_pause)

            else:
                self.logger.info("Using input source position of %.2f and"
                                 "source width of %.2f" %
                                 (self.action.args.cont_source_pos,
                                  self.action.args.cont_source_width))
                auto_mask_type = "UserCont"
                auto_cont_pos = self.action.args.cont_source_pos
                auto_cont_width = self.action.args.cont_source_width

                # First define source extent
                con_pos_mask_0 = auto_cont_pos - auto_cont_width
                con_pos_mask_1 = auto_cont_pos + auto_cont_width

                # Next define lower and upper windows
                con_pos_mask_lo_0 = con_pos_mask_0 - \
                    14 / self.action.args.xbinsize
                con_pos_mask_up_1 = con_pos_mask_1 + \
                    14 / self.action.args.xbinsize

            self.logger.info("Masking all but sky region between "
                             "%.2f and %.2f and between %.2f and %.2f" %
                             (con_pos_mask_lo_0, con_pos_mask_0,
                              con_pos_mask_1, con_pos_mask_up_1))

            # Mask all but local sky from sky calculation
            for i, v in enumerate(binary_mask.flat):
                if (0 < posmap.data.flat[i] < con_pos_mask_lo_0) or \
                   (con_pos_mask_0 < posmap.data.flat[i] < con_pos_mask_1) or \
                   (con_pos_mask_up_1 < posmap.data.flat[i] < posmax):
                    binary_mask.flat[i] = True

        # count masked pixels
        tmsk = len(np.nonzero(np.where(binary_mask.flat, True, False))[0])
        self.logger.info("Number of pixels masked = %d" % tmsk)

        finiteflux = np.isfinite(self.action.args.ccddata.data.flat)

        # get un-masked points mapped to exposed regions on CCD
        # handle dichroic bad region
        if self.action.args.dich:
            if self.action.args.camera == 0:    # Blue
                q = [i for i, v in enumerate(slicemap.data.flat)
                     if 0 <= v <= 23 and
                     posbuf < posmap.data.flat[i] < (posmax - posbuf) and
                     waveall0 <= wavemap.data.flat[i] <= waveall1 and
                     not (v > 20 and wavemap.data.flat[i] > 5600.) and
                     finiteflux[i] and not binary_mask.flat[i]]
            else:                               # Red
                q = [i for i, v in enumerate(slicemap.data.flat)
                     if 0 <= v <= 23 and
                     posbuf < posmap.data.flat[i] < (posmax - posbuf) and
                     waveall0 <= wavemap.data.flat[i] <= waveall1 and
                     not (v > 20 and wavemap.data.flat[i] < 5600.) and
                     finiteflux[i] and not binary_mask.flat[i] and
                     50 <= ymap.data.flat[i] <= (ny - 50)]
        else:
            if self.action.args.camera == 0:    # Blue
                q = [i for i, v in enumerate(slicemap.data.flat)
                     if 0 <= v <= 23 and
                     posbuf < posmap.data.flat[i] < (posmax - posbuf) and
                     waveall0 <= wavemap.data.flat[i] <= waveall1 and
                     finiteflux[i] and not binary_mask.flat[i]]
            else:
                q = [i for i, v in enumerate(slicemap.data.flat)
                     if 0 <= v <= 23 and
                     posbuf < posmap.data.flat[i] < (posmax - posbuf) and
                     waveall0 <= wavemap.data.flat[i] <= waveall1 and
                     finiteflux[i] and not binary_mask.flat[i] and
                     50 <= ymap.data.flat[i] <= (ny - 50)]

        # get all points mapped to exposed regions on the CCD (for output)
        qo = [i for i, v in enumerate(slicemap.data.flat)
              if 0 <= v <= 23 and posmap.data.flat[i] >= 0 and
              waveall0 <= wavemap.data.flat[i] <= waveall1 and
              finiteflux[i]]

        # extract relevant image values
        fluxes = self.action.args.ccddata.data.flat[q]

        # relevant wavelengths
        waves = wavemap.data.flat[q]
        self.logger.info("Number of fit waves = %d" % len(waves))

        # keep output wavelengths
        owaves = wavemap.data.flat[qo]
        self.logger.info("Number of output waves = %d" % len(owaves))

        # sort on wavelength
        s = np.argsort(waves)
        waves = waves[s]
        fluxes = fluxes[s]

        # knots per pixel
        knotspp = self.config.instrument.KNOTSPP
        n = int(sm_sz[0] * knotspp)

        # calculate break points for b splines
        bkpt = np.min(waves) + np.arange(n+1) * \
            (np.max(waves) - np.min(waves)) / n

        # log
        self.logger.info("Nknots = %d, min = %.2f, max = %.2f (A)" %
                         (n, np.min(bkpt), np.max(bkpt)))

        # do bspline fit
        sft0, gmask = Bspline.iterfit(waves, fluxes, fullbkpt=bkpt,
                                      upper=1, lower=1)
        gp = [i for i, v in enumerate(gmask) if v]
        yfit1, _ = sft0.value(waves)
        self.logger.info("Number of good points = %d" % len(gp))

        # check result
        if np.max(yfit1) < 0:
            self.logger.warning("B-spline failure")
            if n > 2000:
                if n == 5000:
                    n = 2000
                if n == 8000:
                    n = 5000
                # calculate breakpoints
                bkpt = np.min(waves) + np.arange(n + 1) * \
                    (np.max(waves) - np.min(waves)) / n
                # log
                self.logger.info("Nknots = %d, min = %.2f, max = %.2f (A)" %
                                 (n, np.min(bkpt), np.max(bkpt)))
                # do bspline fit
                sft0, gmask = Bspline.iterfit(waves, fluxes, fullbkpt=bkpt,
                                              upper=1, lower=1)
                yfit1, _ = sft0.value(waves)
            if np.max(yfit1) <= 0:
                self.logger.warning("B-spline final failure, sky is zero")

        # get values at original wavelengths
        yfit, _ = sft0.value(owaves)

        # for plotting
        gwaves = waves[gp]
        gfluxes = fluxes[gp]
        npts = len(gwaves)
        stride = int(npts / 8000.)
        xplt = gwaves[::stride]
        yplt = gfluxes[::stride]
        fplt, _ = sft0.value(xplt)
        yrng = [np.min(yplt), np.max(yplt)]
        self.logger.info("Stride = %d" % stride)

        # plot, if requested
        if self.config.instrument.plot_level >= 1:
            # output filename stub
            skyfnam = "sky_%05d_%s_%s_%s" % \
                     (self.action.args.ccddata.header['FRAMENO'],
                      self.action.args.illum, self.action.args.grating,
                      self.action.args.ifuname)
            p = figure(
                title=self.action.args.plotlabel + ' Master Sky',
                x_axis_label='Wave (A)',
                y_axis_label='Flux (e-)',
                plot_width=self.config.instrument.plot_width,
                plot_height=self.config.instrument.plot_height)
            p.circle(xplt, yplt, size=1, line_alpha=0., fill_color='purple',
                     legend_label='Data')
            p.line(xplt, fplt, line_color='red', legend_label='Fit')
            p.line([wavegood0, wavegood0], yrng, line_color='green')
            p.line([wavegood1, wavegood1], yrng, line_color='green')
            p.y_range.start = yrng[0]
            p.y_range.end = yrng[1]
            bokeh_plot(p, self.context.bokeh_session)
            if self.config.instrument.plot_level >= 2:
                input("Next? <cr>: ")
            else:
                time.sleep(self.config.instrument.plot_pause)
            save_plot(p, filename=skyfnam+".png")

        # create sky image
        sky = np.zeros(self.action.args.ccddata.data.shape, dtype=float)
        sky.flat[qo] = yfit

        # store original data, header
        img = self.action.args.ccddata.data
        hdr = self.action.args.ccddata.header.copy()
        self.action.args.ccddata.data = sky

        # get master sky output name
        ofn_full = self.action.args.name
        ofn = os.path.basename(ofn_full)
        msname = strip_fname(ofn) + '_' + suffix + '.fits'

        log_string = MakeMasterSky.__module__
        self.action.args.ccddata.header['IMTYPE'] = 'SKY'
        self.action.args.ccddata.header['HISTORY'] = log_string
        self.action.args.ccddata.header['SKYMODEL'] = (True, 'sky model image?')
        self.action.args.ccddata.header['SKYIMAGE'] = \
            (ofn, 'image used for sky model')
        if tmsk > 0:
            self.action.args.ccddata.header['SKYMSK'] = (True,
                                                         'was sky masked?')
            if auto_masked:
                self.action.args.ccddata.header['AUTOMASK'] = (True,
                                                               'auto-masked?')
                self.action.args.ccddata.header['AUTMSKTY'] = (
                    auto_mask_type, 'Type of auto-masking')
                self.action.args.ccddata.header['CONTPOS'] = (
                    auto_cont_pos, 'Position in slice of continuum')
                self.action.args.ccddata.header['CONTWID'] = (
                    auto_cont_width, 'Width of continuum source')
            # self.action.args.ccddata.header['SKYMSKF'] = (skymf,
            # 'sky mask file')
        else:
            self.action.args.ccddata.header['SKYMSK'] = (False,
                                                         'was sky masked?')
        self.action.args.ccddata.header['WAVMAPF'] = wmf
        self.action.args.ccddata.header['SLIMAPF'] = slf
        self.action.args.ccddata.header['POSMAPF'] = pof

        # output master sky
        kcwi_fits_writer(self.action.args.ccddata, output_file=msname,
                         output_dir=self.config.instrument.output_directory)
        self.context.proctab.update_proctab(frame=self.action.args.ccddata,
                                            suffix=suffix,
                                            newtype="SKY",
                                            filename=self.action.args.name)
        self.context.proctab.write_proctab(tfil=self.config.instrument.procfile)

        # restore original image
        self.action.args.ccddata.data = img
        self.action.args.ccddata.header = hdr

        self.logger.info(log_string)
        return self.action.args

    # END: class MakeMasterSky()



from keckdrpframework.primitives.base_primitive import BasePrimitive
from kcwidrp.primitives.kcwi_file_primitives import read_table, kcwi_fits_reader
from kcwidrp.core.bokeh_plotting import bokeh_plot
from kcwidrp.primitives.kcwi_file_primitives import strip_fname, plotlabel

import numpy as np
from numpy.polynomial import polynomial as poly
from scipy.signal import savgol_filter
import os
from skimage import transform as tf
from bokeh.plotting import figure


class ExtractArcs(BasePrimitive):
    """
    Use derived traces to extract arc spectra along continuum bars.

    Reads in traces from continuum bars and then uses them to extract spectra
    along each trace.  Also performs a background subtraction for each extracted
    spectrum.  Adds a HISTORY record to the FITS header.

    Uses the following configuration parameter:

        * saveintims: if set to ``True`` write out a warped version of arc image in \*_warped.fits.  Defaults to ``False``.

    Adds extracted arcs to returned arguments.

    """

    def __init__(self, action, context):
        BasePrimitive.__init__(self, action, context)
        self.logger = context.pipeline_logger
        self.action.args.reference_bar_separation = None
        self.action.args.contbar_image_number = None
        self.action.args.contbar_image = None
        self.action.args.arc_number = None
        self.action.args.arc_image = None
        self.action.args.source_control_points = None
        self.action.args.destination_control_points = None
        self.action.args.bar_id = None
        self.action.args.slice_id = None

    def _pre_condition(self):
        marcs_in_proctable = self.context.proctab.search_proctab(
            frame=self.action.args.ccddata, target_type='MARC',
            nearest=True
        )
        self.logger.info("%d master arc frames found" %
                         len(marcs_in_proctable))
        contbars_in_proctable = self.context.proctab.search_proctab(
            frame=self.action.args.ccddata, target_type='MCBARS',
            nearest=True)
        self.logger.info("%d master continuum bars frames found" %
                         len(contbars_in_proctable))
        if len(contbars_in_proctable) > 0 and len(marcs_in_proctable) > 0:
            self.action.args.original_filename = strip_fname(
                contbars_in_proctable['filename'][0]) + "_trace.fits"
            return True
        else:
            self.action.args.original_filename = None
            return False

    def _perform(self):
        do_plot = (self.config.instrument.plot_level >= 3)
        plab = plotlabel(self.action.args)
        self.logger.info("Extracting arc spectra")
        # Double check
        if not self.action.args.original_filename:
            self.logger.error("No traces found")
            return self.action.args
        # All is ready
        original_filename = self.action.args.original_filename
        self.logger.info("Trace table found: %s" % original_filename)
        # trace = read_table(tab=tab, indir='redux', suffix='trace')
        # Find  and read control points from continuum bars
        trace = read_table(
            input_dir=os.path.join(self.config.instrument.cwd,
                                   self.config.instrument.output_directory),
            file_name=original_filename)
        self.context.trace = {}
        for key in trace.meta.keys():
            self.context.trace[key] = trace.meta[key]
        middle_row = self.context.trace['MIDROW']
        window = self.context.trace['WINDOW']
        self.action.args.reference_bar_separation = self.context.trace[
            'REFDELX']
        self.action.args.contbar_image_number = self.context.trace['CBARSNO']
        self.action.args.contbar_image = self.context.trace['CBARSFL']
        self.action.args.arc_number = self.action.args.ccddata.header['FRAMENO']
        self.action.args.arc_image = self.action.args.ccddata.header['OFNAME']

        self.action.args.source_control_points = trace['src']
        self.action.args.destination_control_points = trace['dst']
        self.action.args.bar_id = trace['barid']
        self.action.args.slice_id = trace['slid']

        self.logger.info("Fitting spatial control points")

        # This is strictly for AIT data!
        if self.config.instrument.NBARS < 60:
            # get trimmed bars image so geometry is consistent
            barsfn = self.action.args.contbar_image.split(
                '.fits')[0] + '_int.fits'
            barim, bartab = kcwi_fits_reader(
                os.path.join(self.config.instrument.output_directory, barsfn))
            arcs = []
            bars = []
            for ib in range(self.config.instrument.NBARS):
                bind = np.where(trace['barid'] == ib)
                xp = trace['dst'][bind][:, 0]
                yp = trace['dst'][bind][:, 1]
                sind = np.argsort(yp)
                yps = yp[sind]
                xps = xp[sind]
                c, stats = poly.polyfit(yps, xps, deg=3, full=True)
                arc = []
                bar = []
                bkg = []
                for iy in range(self.action.args.ccddata.header['NAXIS2']):
                    xv = int(poly.polyval(iy, c))
                    yy = self.action.args.ccddata.data[
                         iy, (xv - window):(xv + window + 1)]
                    byy = barim.data[iy, (xv - window):(xv + window + 1)]
                    nyy = len(yy)
                    if nyy > 0:
                        bkg.append(np.nanmin(yy) * float(nyy))
                        arc.append(np.nansum(yy))
                        bar.append(np.nansum(byy))
                    else:
                        bkg.append(0.)
                        arc.append(0.)
                        bar.append(0.)
                arc = np.asarray(arc, dtype=float)
                bar = np.asarray(bar, dtype=float)
                bkg = np.asarray(bkg, dtype=float)
                bkgf = savgol_filter(bkg, 51, 5)
                if do_plot:
                    xp = np.arange(len(arc))
                    p = figure(title=plab + "ARC # %d" % len(arcs),
                               x_axis_label="Y CCD Pixel",
                               y_axis_label="Flux",
                               plot_width=self.config.instrument.plot_width,
                               plot_height=self.config.instrument.plot_height)
                    p.line(xp, arc, legend_label='Arc', color='blue')
                    p.line(xp, bkg, legend_label='Bkg', color='red')
                    p.line(xp, bkgf, legend_label='BkgFit', color='green')
                    bokeh_plot(p, self.context.bokeh_session)
                    q = input("Next? <cr>, q to quit: ")
                    if 'Q' in q.upper():
                        do_plot = False
                arc -= bkgf
                arcs.append(arc)
                bars.append(bar)
        else:
            # fit transform
            # NOTE: we do not need an asymmetric polynomial for this
            # global fit (only for per slice fitting: see SolveGeom.py)
            tform = tf.estimate_transform(
                'polynomial', self.action.args.source_control_points,
                self.action.args.destination_control_points, order=3)

            self.logger.info("Transforming arc image")
            warped_image = tf.warp(self.action.args.ccddata.data, tform)
            # Write warped arcs if requested
            if self.config.instrument.saveintims:
                from kcwidrp.primitives.kcwi_file_primitives import kcwi_fits_writer
                # write out warped image
                self.action.args.ccddata.data = warped_image
                kcwi_fits_writer(
                    self.action.args.ccddata, table=self.action.args.table,
                    output_file=self.action.args.name,
                    output_dir=self.config.instrument.output_directory,
                    suffix="warped")
                self.logger.info("Transformed arcs produced")
            # extract arcs
            self.logger.info("Extracting arcs")
            arcs = []
            bars = []
            # sectors for bkgnd subtraction
            sectors = 16
            for xyi, xy in enumerate(self.action.args.source_control_points):
                if xy[1] == middle_row:
                    xi = int(xy[0]+0.5)
                    arc = np.nanmedian(
                        warped_image[:, (xi - window):(xi + window + 1)],
                        axis=1)
                    # divide spectrum into sectors
                    div = int((len(arc)-100) / sectors)
                    # get minimum for each sector
                    xv = []
                    yv = []
                    for i in range(sectors):
                        try:
                            mi = np.nanargmin(arc[50+i*div:50+(i+1)*div])
                            mn = np.nanmin(arc[50+i*div:50+(i+1)*div])
                            xv.append(mi+50+i*div)
                            yv.append(mn)
                        except ValueError:
                            self.logger.warning("Bad sector %d" % i)
                    # fit minima to model background
                    res = np.polyfit(xv, yv, 3)
                    xp = np.arange(len(arc))
                    bkg = np.polyval(res, xp)   # resulting model
                    # plot if requested
                    if do_plot:
                        p = figure(
                            title=plab + "ARC # %d" % len(arcs),
                            x_axis_label="Y CCD Pixel", y_axis_label="Flux",
                            plot_width=self.config.instrument.plot_width,
                            plot_height=self.config.instrument.plot_height)
                        p.line(xp, arc, legend_label='Arc', color='blue')
                        p.line(xp, bkg, legend_label='Bkg', color='red')
                        bokeh_plot(p, self.context.bokeh_session)
                        q = input("Next? <cr>, q to quit: ")
                        if 'Q' in q.upper():
                            do_plot = False
                    # subtract model background
                    arc -= bkg
                    # add to arcs list
                    arcs.append(arc)
        # Did we get the correct number of arcs?
        if len(arcs) == self.config.instrument.NBARS:
            self.logger.info("Extracted %d arcs" % len(arcs))
            self.context.arcs = arcs
            if self.config.instrument.NBARS < 60:
                self.context.bars = bars
        else:
            self.logger.error("Did not extract %d arcs, extracted %d" %
                              (self.config.instrument.NBARS, len(arcs)))

        log_string = ExtractArcs.__module__
        self.action.args.ccddata.header['HISTORY'] = log_string
        self.logger.info(log_string)

        return self.action.args
# END: class ExtractArcs()



#!/usr/bin/env python

"""
This script scans a given directory and returns a report on whether the cals
inside the directory meet the minimum requirements needed to reduce the science
frames found inside. It follows the following logic:

#. Finds all OBJECT frames. It uses these to determine what instrument configurations are needed. This is saved in an internal "Proc table"

#. Searches for BIAS and CONTBARS frames for needed setups (1x1 and 2x2)

#. Searches for ARCS and FLATS

#. Searches for matching standard stars for the setup

"""

from pathlib import Path
import argparse
import logging
import warnings
import pkg_resources

from astropy.nddata import CCDData
from astropy.utils.exceptions import AstropyWarning

from kcwidrp.core.kcwi_proctab import Proctab
from keckdrpframework.config.framework_config import ConfigClass
from kcwidrp.core.kcwi_get_std import kcwi_get_std


warnings.simplefilter('ignore', category=AstropyWarning)


def parse_args():
    """Parse arguments passed into this script from the command line

    Returns
    -------
    argparse
        Dict-like object with parsed args

    :meta private:
    """

    parser = argparse.ArgumentParser(
        description="Checks a directory of data to see if the OBJECTs within "
                    "have the required cals.")

    parser.add_argument('filepaths', help="Files to inspect", nargs="+")
    parser.add_argument('-v', '--verbose', dest="verbose", action="store_true",
                        help="Print exhaustive information")
    parser.add_argument('-c', '--config', dest="config", type=str,
                        help="KCWI configuration file", default=None)

    return parser.parse_args()


def check_cal_type(proctab, ccd_frame, setup_frame, targ_type, minimum, logger):
    """Checks the given proctab for calibrations matching the input frame

    Parameters
    ----------
    proctab : Proctab
        Proccessing table instance to search through
    ccd_frame : CCDData
        Astropy CCDData (or kcwidrp KCCDData) object we want to match cals to
    setup_frame : astropy.Table row
        Proctab row corresponding to an OBJECT setup we are looking for
    targ_type : str
        Type of calibration being looked for
    minimum : int
        Minimum number of this type of cal required
    logger : logging.Logger
        Logger for debuging

    Returns
    -------
    str
        String representing the results of the search. PASSED if enough matching cals were found, FAILED otherwise
    
    :meta private:
    """

    found_list = proctab.search_proctab(frame=ccd_frame, target_type=targ_type) #target_group=setup_frame['GRPID']
    logger.debug(f"For setup {setup_frame['CID']}, found {len(found_list)} {targ_type} frames, need {minimum}")
    if len(found_list) >= minimum:
        return "PASSED"
    else:
        return f"FAILED: found {len(found_list): <3}, needed {minimum: <3}"

def main():

    # Parse arguments
    args = parse_args()


    # Set up logging
    logger = logging.getLogger("Logger")
    logger.addHandler(logging.StreamHandler())

    if args.verbose:
        logger.setLevel("DEBUG")
    else:
        logger.setLevel("INFO")
    

    # Load config
    pkg = 'kcwidrp'
    if args.config is None:
        kcwi_config_file = 'configs/kcwi.cfg'
        kcwi_config_fullpath = pkg_resources.resource_filename(pkg, kcwi_config_file)
        config = ConfigClass(kcwi_config_fullpath, default_section='KCWI')
    else:
        config = ConfigClass(args.config, default_section='KCWI')

    logger.debug(f"Loaded KCWI config")

    # Prepare to read files
    files = [Path(file) for file in args.filepaths]
    if len(files) == 0:
        logger.error("No files found! Exiting...")
        return
    logger.info(f"Found {len(files)} files to inspect")


    # Set up proc table. Use a dummy logger, since we don't want its messages
    dummy_logger = logging.getLogger("dummy_logger")
    proctab = Proctab(dummy_logger)
    proctab.read_proctab("temp.proc")


    # Load all files into the proctable
    frames = {}
    for file in files:
        try:
            frame = CCDData.read(file, unit='adu')
        except FileNotFoundError as e:
            logger.error(f"Failed to open {file}")
        proctab.update_proctab(frame, filename=file.name)
        frames[Path(file).name] = frame

    # Get all the unique OBJECT setups:
    objects = proctab.proctab[proctab.proctab["TYPE"] == "OBJECT"]

    standards = {}
    CIDs = set()
    DIDs = set()
    unique_setups = []
    for obj in objects:
        # See if the setup is unique
        if obj["CID"] not in CIDs or obj["DID"] not in DIDs:
            CIDs.add(obj["CID"])
            DIDs.add(obj["DID"])
            unique_setups.append(obj)
        
        # While we're here, compile a dict with all standard stars ordered by CID for later
        stdfile, _ = kcwi_get_std(obj["TARGNAME"], logger=dummy_logger)
        if stdfile is not None:
            CID = standards.get(obj["CID"], None)
            if CID is None:
                standards[obj["CID"]] = {"files" : [], "names" : []}
                CID = standards[obj["CID"]]
            CID["files"].append(obj['filename'])
            CID["names"].append(obj['TARGNAME'])

    # Log results
    logger.info(f"\nFound {len(unique_setups)} individual setups:")

    logger.info(f"{'Config ID': <25}{'Detector ID': <15}{'Camera': <10}{'IFU': <10}{'Grating': <10}{'Grating Angle': <15}{'Central Wavelength': <20}{'Binning': <10}{'Standards': <25}")
    for obj in unique_setups:
        matching_standards = standards.get(obj["CID"], None)
        if matching_standards:
            standards_str = ", ".join(set(matching_standards["names"]))
        else:
            standards_str = "NA"
        logger.info(f"{obj['CID']: <25}{obj['DID']: <15}{obj['CAM']: <10}{obj['IFU']: <10}{obj['GRAT']: <10}{obj['GANG']: <15}{obj['CWAVE']: <20}{obj['BIN']: <10}{standards_str: <25}")


    # Check cals for each setup:
    report = {}
    for setup_frame in unique_setups:
        frame_config = config[setup_frame['CAM']]

        bias_min_frames = int(frame_config['bias_min_nframes'])
        cont_min_frames = int(frame_config['contbars_min_nframes'])
        arc_min_frames = int(frame_config['arc_min_nframes'])
        flat_min_frames = int(frame_config['flat_min_nframes'])


        ccd_frame = frames[setup_frame['filename']]
        results = {
            "BIAS" : "UNCHECKED",
            "CONTBARS" : "UNCHECKED",
            "ARCS" : "UNCHECKED",
            "FLATS" : "UNCHECKED",
            "all_pass": False,
            "STANDARDS" : "UNCHECKED"
        }
        results["BIAS"] = check_cal_type(proctab, ccd_frame, setup_frame, "BIAS", bias_min_frames, logger)
        results["CONTBARS"] = check_cal_type(proctab, ccd_frame, setup_frame, "CONTBARS", cont_min_frames, logger)
        results["ARCS"] = check_cal_type(proctab, ccd_frame, setup_frame, "ARCLAMP", arc_min_frames, logger)
        results["FLATS"] = check_cal_type(proctab, ccd_frame, setup_frame, "FLATLAMP", flat_min_frames, logger)
        
        # Check the standards list from earlier for matching setups
        matching_standards = standards.get(setup_frame["CID"], None)
        if matching_standards is not None:
            out = {}
            for (name, file) in zip(matching_standards["names"], matching_standards["files"]):
                if out.get(name, None):
                    out[name].append(file)
                else:
                    out[name] = [file]
            out_str = ""
            for name in out.keys():
                out_str += f"{name} ({', '.join(out[name])})"
            # results["STANDARDS"] = ", ".join(matching_standards["names"])
            results["STANDARDS"] = out_str
            standards_result = True
        else:
            results["STANDARDS"] = "FAILED"
            standards_result = False
        
        # Collate a final "are we good here" boolean for the report
        results["all_pass"] = bool(results["BIAS"] == "PASSED" and 
                            results["CONTBARS"] == "PASSED" and
                            results["ARCS"] == "PASSED" and
                            results["FLATS"] == "PASSED" and
                            standards_result)
        
        report[setup_frame['CID']] = results


    # Print results, passes then fails
    passes = []
    fails = []
    logger.info("")
    
    # Collect the passes and fails into their own lists
    for setup_id in report.keys():
        if report[setup_id]['all_pass']:
            passes.append(setup_id)
        else:
            fails.append(setup_id)
    
    # Log info on passes (in green text)
    if len(passes) > 0:
        logger.info("Following setups \033[32mPASSED\033[0m:")
        logger.info(", ".join(passes))
        if args.verbose:
            for p in passes:
                logger.debug(f"Setup {p} will use the following standard stars:")
                logger.debug(report[p]["STANDARDS"])
    
    # Log info on all the fails (in red text)
    if len(fails) > 0:
        logger.info("Following setups \033[31mFAILED\033[0m:")
        for fail in fails:
            logger.info(f"{fail}:")
            for key in report[fail].keys():
                if key == "all_pass" : continue
                logger.info(f"\t{key: <10}\t{report[fail][key]: <20}")
            logger.info("\n\tThis effects the following OBJECT frames:")
            logger.info('\t' + str(objects[objects["CID"] == fail]['filename', 'TARGNAME']).replace('\n', '\n\t'))
    else:
        logger.info("\033[32mNo failures to report.\033[0m")

if __name__ == "__main__":
    main()


from keckdrpframework.models.arguments import Arguments
from keckdrpframework.primitives.base_img import BaseImg
from kcwidrp.primitives.kcwi_file_primitives import kcwi_fits_reader, \
    kcwi_fits_writer, parse_imsec, strip_fname
from kcwidrp.core.bokeh_plotting import bokeh_plot
from kcwidrp.core.kcwi_plotting import save_plot

from bokeh.plotting import figure
import ccdproc
import numpy as np
from scipy.stats import sigmaclip
from astropy.stats import mad_std
import time
import os


class MakeMasterBias(BaseImg):
    """
    Stack bias frames into a master bias frame.

    Generate a master bias image from overscan-subtracted and trimmed bias
    frames (\*_intb.fits) based on the instrument config parameter
    bias_min_nframes, which defaults to 7.  The combine method for biases is
    'average' and so cosmic rays may be present, especially in RED channel data.
    A high sigma clipping of 2.0 is used to help with the CRs.

    Uses the ccdproc.combine routine to perform the stacking.

    Writes out a \*_mbias.fits file and records a master bias frame in the proc
    table.


    """

    def __init__(self, action, context):
        BaseImg.__init__(self, action, context)
        self.logger = context.pipeline_logger

    def _pre_condition(self):
        """
        Checks if we can build a stacked frame based on the processing table
        """
        # Get bias count
        self.logger.info("Checking precondition for MakeMasterBias")
        self.combine_list = self.context.proctab.search_proctab(
            frame=self.action.args.ccddata, target_type='BIAS',
            target_group=self.action.args.groupid)
        self.logger.info(f"pre condition got {len(self.combine_list)},"
                         f" expecting {self.action.args.min_files}")
        # Did we meet our pre-condition?
        if len(self.combine_list) >= self.action.args.min_files:
            return True
        else:
            return False

    def _perform(self):
        """
        Returns an Argument() with the parameters that depends on this operation
        """
        method = 'average'
        sig_up = 2.0        # default upper sigma rejection limit
        suffix = self.action.args.new_type.lower()

        combine_list = list(self.combine_list['filename'])
        # get master bias output name
        # mbname = combine_list[-1].split('.fits')[0] + '_' + suffix + '.fits'
        mbname = strip_fname(combine_list[0]) + '_' + suffix + '.fits'
        # mbname = master_bias_name(self.action.args.ccddata)
        bsec, dsec, tsec, direc, amps, aoff = self.action.args.map_ccd

        # loop over amps
        stack = []
        stackf = []
        for bias in combine_list:
            inbias = bias.split('.fits')[0] + '_intb.fits'
            stackf.append(inbias)
            # using [0] drops the table and leaves just the image
            stack.append(kcwi_fits_reader(
                os.path.join(self.context.config.instrument.cwd, 'redux',
                             inbias))[0])

        stacked = ccdproc.combine(stack, method=method, sigma_clip=True,
                                  sigma_clip_low_thresh=None,
                                  sigma_clip_high_thresh=sig_up,
                                  sigma_clip_func=np.ma.median,
                                  sigma_clip_dev_func=mad_std)
        stacked.header['IMTYPE'] = self.action.args.new_type
        stacked.header['NSTACK'] = (len(combine_list),
                                    'number of images stacked')
        stacked.header['STCKMETH'] = (method, 'method used for stacking')
        stacked.header['STCKSIGU'] = (sig_up,
                                      'Upper sigma rejection for stacking')
        for ii, fname in enumerate(stackf):
            fname_base = os.path.basename(fname)
            stacked.header['STACKF%d' % (ii + 1)] = (fname_base,
                                                     "stack input file")

        # for readnoise stats use 2nd and 3rd bias
        diff = stack[1].data.astype(np.float32) - \
            stack[2].data.astype(np.float32)
        namps = stack[1].header['NVIDINP']
        if len(amps) != namps:
            self.logger.warning("Amp count disagreement!")
        for ia in amps:
            # get gain
            gain = stacked.header['GAIN%d' % ia]
            # get amp section
            sec, rfor = parse_imsec(stacked.header['ATSEC%d' % ia])
            noise = diff[sec[0]:(sec[1]+1), sec[2]:(sec[3]+1)]
            noise = np.reshape(noise, noise.shape[0]*noise.shape[1]) * \
                gain / 1.414
            # get stats on noise
            c, low, upp = sigmaclip(noise, low=3.5, high=3.5)
            bias_rn = c.std()
            self.logger.info("Amp%d read noise from bias in e-: %.3f" %
                             (ia, bias_rn))
            stacked.header['BIASRN%d' % ia] = \
                (float("%.3f" % bias_rn), "RN in e- from bias")
            if self.config.instrument.plot_level >= 1:
                # output filename stub
                biasfnam = "bias_%05d_amp%d_rdnoise" % \
                          (stacked.header['FRAMENO'], ia)
                plabel = '[ Img # %d' % stacked.header['FRAMENO']
                plabel += ' (Bias)'
                plabel += ' %s' % self.action.args.ccddata.header['BINNING']
                plabel += ' %s' % self.action.args.ccddata.header['AMPMODE']
                plabel += ' %d' % self.action.args.ccddata.header['GAINMUL']
                plabel += ' %s' % ('fast' if
                                   self.action.args.ccddata.header['CCDMODE']
                                   else 'slow')
                plabel += ' ] '
                hist, edges = np.histogram(noise, range=(low, upp),
                                           density=False, bins=50)
                x = np.linspace(low, upp, 500)
                pdf = np.max(hist)*np.exp(-x**2/(2.*bias_rn**2))
                p = figure(title=plabel+'BIAS NOISE amp %d = %.3f' %
                           (ia, bias_rn),
                           x_axis_label='e-', y_axis_label='N',
                           plot_width=self.config.instrument.plot_width,
                           plot_height=self.config.instrument.plot_height)
                p.quad(top=hist, bottom=0, left=edges[:-1], right=edges[1:],
                       fill_color="navy", line_color="white", alpha=0.5)
                p.line(x, pdf, line_color="#ff8888", line_width=4, alpha=0.7,
                       legend_label="PDF")
                p.line([-bias_rn, -bias_rn], [0, np.max(hist)], color='red',
                       legend_label="Sigma")
                p.line([bias_rn, bias_rn], [0, np.max(hist)], color='red')
                p.y_range.start = 0
                bokeh_plot(p, self.context.bokeh_session)
                if self.config.instrument.plot_level >= 2:
                    input("Next? <cr>: ")
                else:
                    time.sleep(self.config.instrument.plot_pause)
                save_plot(p, filename=biasfnam+".png")

        log_string = MakeMasterBias.__module__
        stacked.header['HISTORY'] = log_string
        self.logger.info(log_string)

        kcwi_fits_writer(stacked, output_file=mbname,
                         output_dir=self.config.instrument.output_directory)
        self.context.proctab.update_proctab(frame=stacked, suffix=suffix,
                                            newtype=self.action.args.new_type,
                                            filename=stacked.header['OFNAME'])
        self.context.proctab.write_proctab(tfil=self.config.instrument.procfile)
        return Arguments(name=mbname)
    # END: class ProcessBias()
